//
// File:      cam.C
// Author:    Xing Cindy Chen <xcchen@graphics.stanford.edu>
//
// Description:  routines related camera geometry
//
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fstream.h>
#include <math.h>
#include "CAM.h"

//==========================================================================
void CAM::
rotMatrix2Euler(double r[3][3],double& anglex,double& angley,double& anglez)
{
  double eps=0.0000005;

	     
  angley = atan2(-r[2][0],sqrt(r[0][0]*r[0][0]+r[1][0]*r[1][0]));

  if (angley < M_PI/2+eps && angley >M_PI/2-eps) {
    angley = M_PI/2;
    anglez = 0.0;
    anglex = atan2(r[0][1],r[1][1]);
  } else if (angley < -M_PI/2+eps && angley >-M_PI/2-eps) {
    angley = -M_PI/2;
    anglez = 0.0;
    anglex = -atan2(r[0][1],r[1][1]);
  } else {
    double  cbinv=1.0/cos(angley);
    anglez = atan2(r[1][0]*cbinv, r[0][0]*cbinv);
    anglex = atan2(r[2][1]*cbinv, r[2][2]*cbinv);
  }

  //  printf("angle xyz: %f %f %f\n",anglex,angley,anglez);
  
}

void CAM::
error(char* s, char* s2 )
{
  cerr << s << ' ' << s2 << '\n';
  exit(1);
}


void CAM::getCamCenter(double* camCenter) const
{
  camCenter[0]=C.x;
  camCenter[1]=C.y;
  camCenter[2]=C.z;
}

void CAM::getCamSelfEulerAngle(double *angle) const
{
  angle[0]=-Rx;
  angle[1]=-Ry;
  angle[2]=-Rz;
}

void CAM::getOpAxis(double *opaxis)
{
  /* we just want to know the coordinates in world frame
  of vector (0,0,1) in camera frame, which is the third column
  of the rotation matrix from camera frame to world frame, which
  is basicly the 3rd ROW of the stored rotation matrix which converts
  from world to camera frame. */
  opaxis[0]=r7;
  opaxis[1]=r8;
  opaxis[2]=r9;
}

void CAM::getUpVector(double *up)
{
  //up vector is -y axis of camera 
  up[0]=-r4;
  up[1]=-r5;
  up[2]=-r6;
}

void CAM::getIntrinsicMatrix(double A[3][3]) const
{
 
  A[0][0]=f*Asp/dpx;
  A[0][1]=0.0;
  A[0][2]= Cx;
  A[1][0]=0.0;
  A[1][1]= f/dpy;
  A[1][2]= Cy;
  A[2][0]=0.0;
  A[2][1]=0.0;
  A[2][2]=1.0;
}

void CAM::getProjMatrix(double *Proj) const
{
#if 0
  Proj[0]=P.r(1,1);
  Proj[1]=P.r(2,1);
  Proj[2]=P.r(3,1);
  Proj[3]=P.r(1,2);
  Proj[4]=P.r(2,2);
  Proj[5]=P.r(3,2);
  Proj[6]=P.r(1,3);
  Proj[7]=P.r(2,3);
  Proj[8]=P.r(3,3);
  
  Proj[9]=P.r(1,4);
  Proj[10]=P.r(2,4);
  Proj[11]=P.r(3,4);
#else
  
  Proj[0]=P.T[0][0];
  Proj[1]=P.T[1][0];
  Proj[2]=P.T[2][0];
  Proj[3]=P.T[0][1];
  Proj[4]=P.T[1][1];
  Proj[5]=P.T[2][1];
  Proj[6]=P.T[0][2];
  Proj[7]=P.T[1][2];
  Proj[8]=P.T[2][2];
  
  Proj[9]=P.translation.x;
  Proj[10]=P.translation.y;
  Proj[11]=P.translation.z;

#endif
}

void CAM::getGLModelViewMatrix(double *modelView) const
{
  modelView[0]=r1;
  modelView[1]=r4;
  modelView[2]=r7;
  modelView[3]=0.0;
  modelView[4]=r2;
  modelView[5]=r5;
  modelView[6]=r8;
  modelView[7]=0.0;
  modelView[8]=r3;
  modelView[9]=r6;
  modelView[10]=r9;
  modelView[11]=0.0;
  modelView[12]=Tx;
  modelView[13]=Ty;
  modelView[14]=Tz;
  modelView[15]=1.0;

}


void CAM::
getGLFrustumParams(int imXsize, int imYsize, double near,
		   double &left,double &right,double &bottom,double &top) const
{
  double dpx_over_focal=dpx/f;
  double dpy_over_focal=dpy/f;
  
  left= -near*dpx_over_focal/Asp*Cx;
  right= near*dpx_over_focal/Asp*(imXsize-Cx);
  bottom= -near*dpy_over_focal*(imYsize-Cy);
  top=near*dpy_over_focal*Cy;
}

void CAM::setPos(double Tx_, double Ty_, double Tz_)
{
  Tx=Tx_;
  Ty=Ty_;
  Tz=Tz_;
#if 0
  Mm translate=(BR(Tx),semi,Ty,semi,Tz);
  Mm intrinsic=(BR(f*Asp/dpx), 0.0,   Cx, semi,
		0.0,           f/dpy, Cy, semi,
		0.0,           0.0,  1.0);
  P(colon(1,1,3),4)=intrinsic*translate;
  //update invP --no change in this case
  //update C
  C=-invP*P(colon(1,1,3),4);
#else
  Vector3 translate(Tx,Ty,Tz);
  Matrix3 intrinsic(f*Asp/dpx, 0.0,   Cx,
		     0.0,           f/dpy, Cy,
		     0.0,           0.0,  1.0);
  P.translation = intrinsic.apply(translate);
  C=-invP.apply(P.translation);
#endif
}



void CAM::setRot(double Rx_, double Ry_, double Rz_)
{
  Rx=Rx_;
  Ry=Ry_;
  Rz=Rz_;
  double sa = sin(Rx);
  double ca = cos(Rx);
  double sb = sin(Ry);
  double cb = cos(Ry);
  double sg = sin(Rz);
  double cg = cos(Rz);

  r1 = cb * cg;
  r2 = cg * sa * sb - ca * sg;
  r3 = sa * sg + ca * cg * sb;
  r4 = cb * sg;
  r5 = sa * sb * sg + ca * cg;
  r6 = ca * sb * sg - cg * sa;
  r7 = -sb;
  r8 = cb * sa;
  r9 = ca * cb;


  
  //update quaternion
  Rquat=QuatFromRowMatrix(rotm);

  //update projection matrix
  updateProjMatrixAndCamCenter();
  

}



void CAM::setRotQuat(Quaternion quat)
{
  Rquat=quat;

  //update rotation matrix
  Rquat.getRowMatrix(rotm);

  //update Euler angles
  rotMatrix2Euler(rotm,Rx,Ry,Rz);
  
  //update projection matrix
  updateProjMatrixAndCamCenter();
}

void CAM::lookAt(double xfrom,double yfrom,double zfrom,
		 double xto,double yto,double zto)
{
  Vector3 vfrom(xfrom,yfrom,zfrom);
  Vector3 vto(xto,yto,zto);

  Vector3 opaxis_old(0.0,0.0,1.0);
  Vector3 opaxis_new=vto-vfrom;
  opaxis_new=opaxis_new.normalize(); //normalize

  //calc the rotation that rotates the optical axis from old to new direction
  Vector3 rotaxis=cross(opaxis_old,opaxis_new);//rotation axis
  double rotangle_sin=rotaxis.length();
  rotaxis = rotaxis/rotangle_sin; //normalized rotation axis
  double rotangle_cos=dot(opaxis_old,opaxis_new);//cosine of rot. angle
  
  double rotangle=acos(rotangle_cos);
  //printf("angle: %lf, axis: %lf %lf %lf\n",rotangle,rotaxis[0],rotaxis[1],rotaxis[2]);
  
  // The rotation defined by (rotangle,rotaxis) rotates the zaxis (0,0,1)
  // to its new direction; it can also interpret as the mapping of a point
  // from camera frame to world frame, but the rotation matrix stored in
  // CAM is the inverse: from world frame to camera
  // So we negate the rotangle to get the inverse rotation
  rotangle=-rotangle;

  // convert (angle, axis) to quaternion
  double q1=cos(rotangle/2);
  double q2=rotaxis.x * sin(rotangle/2);
  double q3=rotaxis.y * sin(rotangle/2);
  double q4=rotaxis.z * sin(rotangle/2);

  //set rotation
  setRotQuat(q1,q2,q3,q4);
  
  // set translation
  setCamCenter(xfrom,yfrom,zfrom);

}

//set the up vector of the camera which is the -y axis of the cam. frame
void CAM::lookAt(double xfrom,double yfrom,double zfrom,
		 double xto,double yto,double zto,
		 double xup,double yup, double zup)
{

  Vector3 vfrom(xfrom,yfrom,zfrom);
  Vector3 vto(xto,yto,zto);

  Vector3 opaxis_new=vto-vfrom;
  opaxis_new=opaxis_new.normalize(); //normalize

  //project desired y axis
  Vector3 newy=-Vector3(xup,yup,zup);
  Vector3 normal_of_plane_newy_opaxis=cross(newy,opaxis_new);
  if (normal_of_plane_newy_opaxis.length()<0.000001) {
    //opaxis and new y parallel, ignore the up direction
    lookAt(xfrom,yfrom,zfrom,xto,yto,zto);
    return;
  }
  
  Vector3 newy_projection=cross(opaxis_new,normal_of_plane_newy_opaxis);
  Vector3 yaxis_new =newy_projection.normalize(); //normalize
  
  Vector3 xaxis_new = cross(yaxis_new, opaxis_new);
  
  Matrix3 new_camera_to_world(xaxis_new.x,yaxis_new.x,opaxis_new.x,
			      xaxis_new.y,yaxis_new.y,opaxis_new.y,
			      xaxis_new.z,yaxis_new.z,opaxis_new.z);
  
  //but rotm converts world to camera, i.e. we need to transpose
  r1=new_camera_to_world.T[0][0];
  r2=new_camera_to_world.T[1][0];
  r3=new_camera_to_world.T[2][0];
  r4=new_camera_to_world.T[0][1];
  r5=new_camera_to_world.T[1][1];
  r6=new_camera_to_world.T[2][1];
  r7=new_camera_to_world.T[0][2];
  r8=new_camera_to_world.T[1][2];
  r9=new_camera_to_world.T[2][2];

  Quaternion quat=QuatFromRowMatrix(rotm);
  
   //set rotation
  setRotQuat(quat);
  
  // set translation
  setCamCenter(xfrom,yfrom,zfrom);

}

void CAM::adjustToRefCamera(CAM& refcam)
{
  //adjust translation first since it needs the old rotation
  //TranslationNew = RotationMatrixOld * TranslationOfRefCamera + TranslationOld
  Tx += r1*refcam.Tx + r2*refcam.Ty + r3*refcam.Tz;
  Ty += r4*refcam.Tx + r5*refcam.Ty + r6*refcam.Tz;
  Tz += r7*refcam.Tx + r8*refcam.Ty + r9*refcam.Tz;

  // adjust rotation quaternion
  // RotMatrixNew = RotMatrixOld * RotMatrixOfRefCamera;
  Rquat=Rquat*refcam.Rquat;
  
  
  //update rotation matrix
  Rquat.getRowMatrix(rotm);
  
    //update Euler angles
  rotMatrix2Euler(rotm,Rx,Ry,Rz);
  
  //update projection matrix
  updateProjMatrixAndCamCenter();
  
}

/* This routine convert from image coord. to world coordinates, 
   given z  value in world coord. */
void CAM::
imageToWorld(double Xfd, double Yfd, double zw,
	     double &xw, double &yw)
{

  /* convert from image to distorted sensor coordinates */
  double Xd = dpx * (Xfd - Cx) / Asp;
  double Yd = dpy * (Yfd - Cy);

  /* convert from distorted sensor to undistorted sensor plane coordinates */
  double Xu, Yu;
  distorted_to_undistorted_sensor_coord (Xd, Yd, Xu, Yu);

  /* calculate the corresponding xw and yw world coordinates   */
  /* (these equations were derived by simply inverting         */
  /* the perspective projection equations using Macsyma)       */
  double common_denominator = ((r1 * r8 - r2 * r7) * Yu +
			       (r5 * r7 - r4 * r8) * Xu -
			       f * r1 * r5 + f * r2 * r4);

  xw = (((r2 * r9 - r3 * r8) * Yu +
	 (r6 * r8 - r5 * r9) * Xu -
	 f * r2 * r6 + f * r3 * r5) * zw +
	(r2 * Tz - r8 * Tx) * Yu +
	(r8 * Ty - r5 * Tz) * Xu -
	f * r2 * Ty + f * r5 * Tx) / common_denominator;

  yw = -(((r1 * r9 - r3 * r7) * Yu +
	  (r6 * r7 - r4 * r9) * Xu -
	  f * r1 * r6 + f * r3 * r4) * zw +
	 (r1 * Tz - r7 * Tx) * Yu +
	 (r7 * Ty - r4 * Tz) * Xu -
	 f * r1 * Ty + f * r4 * Tx) / common_denominator;

}

void CAM::
imageToWorld(double ximage, double yimage,
	     double &xworld, double &yworld)
{
  imageToWorld(ximage,yimage,0,xworld, yworld);
}

void CAM::
worldToCamera(double xworld, double yworld, double zworld,
	      double& xc, double& yc,double& zc)
{
  /* convert from world coordinates to camera coordinates */
  xc = r1 * xworld + r2 * yworld + r3 * zworld + Tx;
  yc = r4 * xworld + r5 * yworld + r6 * zworld + Ty;
  zc = r7 * xworld + r8 * yworld + r9 * zworld + Tz;

}

/***********************************************************************\
* This routine takes the position of a point in world coordinates [mm]  *
* and determines the position of its image in image coordinates [pix].  *
\***********************************************************************/
double CAM::
worldToImage(double xworld, double yworld, double zworld,
	     double &ximage, double &yimage)
{
  double    xc, yc,zc, Xu, Yu,Xd, Yd;

  /* convert from world coordinates to camera coordinates */
  xc = r1 * xworld + r2 * yworld + r3 * zworld + Tx;
  yc = r4 * xworld + r5 * yworld + r6 * zworld + Ty;
  zc = r7 * xworld + r8 * yworld + r9 * zworld + Tz;

  /* convert from camera coordinates to undistorted sensor plane coordinates */
  Xu = f * xc / zc;
  Yu = f * yc / zc;

  /* convert from undistorted to distorted sensor plane coordinates */
  undistorted_to_distorted_sensor_coord (Xu, Yu, Xd, Yd);

  /* convert from distorted sensor plane coordinates to image coordinates */
  ximage = Xd * Asp / dpx + Cx;
  yimage = Yd / dpy + Cy;

  return(zc);
}

double CAM::
worldToUndistortedImage(double xworld, double yworld, double zworld,
			double &ximage, double &yimage)
{
  double    xc, yc,zc, Xu, Yu,Xd, Yd;
  // printf("%.4f  %.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f\n", r1,r2,r3,r4,r5,r6,r7,r8,r9);
  //printf("%f %f %f %f %f %f\n",f,Asp,Cx,Cy,dpx,dpy);
  //printf("%f %f %f %f %f %f\n",Tx,Ty,Tz,Rx,Ry,Rz); 
  
  /* convert from world coordinates to camera coordinates */
  xc = r1 * xworld + r2 * yworld + r3 * zworld + Tx;
  yc = r4 * xworld + r5 * yworld + r6 * zworld + Ty;
  zc = r7 * xworld + r8 * yworld + r9 * zworld + Tz;

  /* convert from camera coordinates to undistorted sensor plane coordinates */
  Xu = f * xc / zc;
  Yu = f * yc / zc;

  /* convert from distorted sensor plane coordinates to image coordinates */
  ximage = Xu * Asp / dpx + Cx;
  yimage = Yu / dpy + Cy;

  return(zc);
}



/************************************************************************/
void  CAM::    
undistorted_to_distorted_image_coord (double Xfu, double Yfu, 
				      double &Xfd, double &Yfd)
{
    double    Xu,
              Yu,
              Xd,
              Yd;

    /* convert from image to sensor coordinates */
    Xu = dpx * (Xfu - Cx) / Asp;
    Yu = dpy * (Yfu - Cy);

    /* convert from undistorted sensor to distorted sensor plane coordinates */
    undistorted_to_distorted_sensor_coord (Xu, Yu, Xd, Yd);

    /* convert from sensor to image coordinates */
    Xfd = Xd * Asp / dpx + Cx;
    Yfd = Yd / dpy + Cy;
}


/************************************************************************/
void  CAM::
distorted_to_undistorted_image_coord (double Xfd, double Yfd, 
				      double &Xfu, double &Yfu)
{
    double    Xd,
              Yd,
              Xu,
              Yu;

    /* convert from image to sensor coordinates */
    Xd = dpx * (Xfd - Cx) / Asp;
    Yd = dpy * (Yfd - Cy);

    /* convert from distorted sensor to undistorted sensor plane coordinates */
    distorted_to_undistorted_sensor_coord (Xd, Yd, Xu, Yu);

    /* convert from sensor to image coordinates */
    Xfu = Xu * Asp / dpx + Cx;
    Yfu = Yu / dpy + Cy;
}



void CAM::imageToLine3d(double ximage0,double yimage0,
		     double *startPoint,double *direction)
{
  //start point is camera center
  startPoint[0]=C.x;
  startPoint[1]=C.y;
  startPoint[2]=C.z;

  double xu0,yu0;
  distorted_to_undistorted_image_coord(ximage0,yimage0,xu0,yu0);
  Vector3 m0(xu0,yu0,1.0); 
  Vector3 V0= invP.apply(m0);  // point at infinity of ray defined by C0 and m0 
  double invlen=1.0/V0.length();
  V0=V0*invlen;
  
  direction[0]=V0.x;
  direction[1]=V0.y;
  direction[2]=V0.z;
}

void CAM::undistortedImageToLine3d(double ximage0,double yimage0,
		     double *startPoint,double *direction)
{
  //start point is camera center
  startPoint[0]=C.x;
  startPoint[1]=C.y;
  startPoint[2]=C.z;

  double xu0(ximage0),yu0(yimage0);
  Vector3 m0(xu0,yu0,1.0); 
  Vector3 V0= invP.apply(m0);  // point at infinity of ray defined by C0 and m0 
  double invlen=1.0/V0.length();
  V0=V0*invlen;
  
  direction[0]=V0.x;
  direction[1]=V0.y;
  direction[2]=V0.z;
}

/*****************************************************************/ 
/* This routine project the 3D line defined by the center of 
   camera cam1 and a 2D image point (ximage1,yimage1) seen by cam1 
   to a 2D line on the image plane2 of "this" camera. 
*/
void CAM::
epipolarLine(CAM &cam1, 
	     double ximage1, double yimage1,
	     double &xbegin, double &ybegin, double &xend, double &yend)
{
  Point3 C1=cam1.C;
  Point3 e2 = P.apply(C1); // projection of camera center C of cam1 on image
		    // plane2

  /* correct barrel distortion of the image point */
  double xu1,yu1;
  cam1.distorted_to_undistorted_image_coord(ximage1,yimage1,xu1,yu1);
  Point3 m1(xu1,yu1,1.0); 
  Vector3 D = cam1.invP.apply(m1); // point at infinity of ray defined by C1 and m1
  double lambda=5000;
  Point3 M = C1+D*lambda;            //a point on the ray
  Point3 m2 = P.apply(M); // projection of M on image plane 2

  xbegin=e2.x/e2.z;
  ybegin=e2.y/e2.z;
  xend=m2.x/m2.z;
  yend=m2.y/m2.z;

  //printf("camera1 at: (%f %f %f)\n",C1.x,C1.y,C1.z);
  //printf("epipolar line: (%f %f),(%f %f)\n",xbegin,ybegin,xend,yend);
 
}

/* Intersects the two epipolar lines in 3D defined by (ximage0,yimage0) of
   this CAM and (ximage1,yimage1) of cam1. The 3D intersection point is
   given by (x3d,y3d,z3d). If the two lines are skewed,  (x3d,y3d,z3d)
   is the mid point that's closest to both lines. point3d() returns 0
   when the two lines are parallel; otherwise it returns 1.
   Ref: Graphics Gems I, pp304.
*/
int CAM::
point3d(double ximage0,double yimage0,
	CAM &cam1, double ximage1, double yimage1,
	double &x3d,double &y3d, double &z3d,double &err)
{
  // camera centers
  Point3 C0=C;
  Point3 C1=cam1.C;

  // direction of epipolar line 0
  double xu0,yu0;
  distorted_to_undistorted_image_coord(ximage0,yimage0,xu0,yu0);
  Point3 m0(xu0,yu0,1.0); 
  Vector3 V0= invP.apply(m0); // point at infinity of ray defined by C0 and m0

  // direction of epipolar line 1
  double xu1,yu1;
  cam1.distorted_to_undistorted_image_coord(ximage1,yimage1,xu1,yu1);
  Point3 m1(xu1,yu1,1.0); 
  Vector3 V1= cam1.invP.apply(m1); // point at infinity of ray defined by C1 and m1

  // The two lines are defined by L1(t)=C1+V1*t,L0(s)=C0+V0*s
  // The intersection occurs when L1(t)=L0(s). We solve for s and t.
  Vector3 V1xV0=cross(V1,V0);
  double len2=V1xV0.length_squared();
  if (len2==0) // two lines parallel
    return 0;
  else { // not parallel
    Vector3 C0_C1=C0-C1;
    Matrix3 Mt(C0_C1.x, V0.x, V1xV0.x,
	       C0_C1.y, V0.y, V1xV0.y,
	       C0_C1.z, V0.z, V1xV0.z);
    Matrix3 Ms(C0_C1.x, V1.x, V1xV0.x,
	       C0_C1.y, V1.y, V1xV0.y,
	       C0_C1.z, V1.z, V1xV0.z);
    double t=Mt.determinant()/len2;
    double s=Ms.determinant()/len2;

    Point3 L1=C1+V1*t;
    Point3 L0=C0+V0*s;

    Point3 midp=midpoint(L1,L0);

    x3d=midp.x;
    y3d=midp.y;
    z3d=midp.z;
    err=distance(midp,L0);
    return 1;
  }
}


#ifdef USEMATLIB
/* This computes world to undistorted image coord (Xi,Yi) and save
   some intermediate results for future use */
double CAM::
worldToUndistortedImage(Mm Pw /* 3x1 */,
                                Mw Xi_Yi_Zc /* 1x3 */,
                                Mw R_world2cam_Pw /* 1x3 */)
{
  //printf("worldToUndistortedImage() starts...\n");
  //rotation matrix
  //disp(Pw);
  //printf("r1: %lf, r9: %lf\n",r1,r9);

  Mm R_world2cam = (BR(r1),r2,r3, semi,
		    r4,r5,r6, semi,
		    r7,r8,r9);

  //disp(R_world2cam);
  // make sure Pw is a row vector
  //if (size(Pw,1) == 1) Pw = ctranspose(Pw);
  
  Mm R_Pw = R_world2cam * Pw;
  //  disp(R_Pw);
  double Xc = Tx + R_Pw.r(1);
  double Yc = Ty + R_Pw.r(2);
  double Zc = Tz + R_Pw.r(3);
  //printf("Xc: %lf, Yc: %lf, Zc: %lf\n",Xc,Yc,Zc);
  double Zc_inv=1.0/Zc;
  double Xu = Xc * Zc_inv;
  double Yu = Yc * Zc_inv;
  double xfactor = f * Asp / dpx;
  double yfactor = f / dpy ;
  
  double Xi = Xu * xfactor + Cx;
  double Yi = Yu * yfactor + Cy;
  //printf("Xi: %lf, Yi: %lf\n",Xi,Yi);
  // return results
  Xi_Yi_Zc = (BR(Xi),Yi,Zc_inv);
  R_world2cam_Pw = R_Pw;
  //printf("before worldToUndistortedImage() return\n");
  return(Zc);
}


void CAM::undistoredImageJacobian(Mm XYZworld /*3x1 vector */,
				double & Xi, double& Yi,
				Mm& H, char flag)
{
  Mm rot_row1=(BR(r1),r2,r3);
  Mm rot_row2=(BR(r4),r5,r6);
  Mm rot_row3=(BR(r7),r8,r9);
  
  //calculate the projected image coord.
  double Xc= Tx + (rot_row1 * XYZworld).r(); 
  double Yc= Ty + (rot_row2 * XYZworld).r();
  double Zc= Tz + (rot_row3 * XYZworld).r();

  double Zc_inv=1.0/Zc;
  double Xu = Xc * Zc_inv;
  double Yu = Yc * Zc_inv;
  double xfactor = f * Asp / dpx;
  double yfactor = f / dpy ;
  
  Xi = Xu * xfactor + Cx;
  Yi = Yu * yfactor + Cy;
  MY_DEBUG(cout<<"Undistorted image coord: ["<<Xi<<","<<Yi<<"]\n");
  
  // jacobian wrt to point
  int col=1;
  if (flag&D_POINT) {
    H(colon(1,1,1),colon(col,1,col+2))= xfactor *(rot_row1-Xu*rot_row3)*Zc_inv; //d_ui_d_pos
    H(colon(2,1,2),colon(col,1,col+2))= yfactor * (rot_row2-Yu*rot_row3)*Zc_inv; //d_vi_d_pos
    MY_DEBUG(cout<<"Jacobian wrt to point:\n");
    MY_DEBUG(disp(H(colon(1,1,2),colon(col,1,col+2))));
    col+=3;
  }

  // jacobain wrt camera position
  if (flag&D_POS) {
    H(colon(1,1,1),colon(col,1,col+2))= (BR(xfactor * Zc_inv), 0.0, -xfactor*Xu*Zc_inv);
    H(colon(2,1,2),colon(col,1,col+2))= (BR(0.0), yfactor * Zc_inv, -yfactor*Yu*Zc_inv);
    MY_DEBUG(cout<<"Jacobian wrt to camera position:\n");
    MY_DEBUG(disp(H(colon(1,1,2),colon(col,1,col+2))));
    col+=3;
  }

  // jacobain wrt camera  rotation
  if (flag & D_ROT) {
    double cx=cos(Rx);
    double sx=sin(Rx);
    double cy=cos(Ry);
    double sy=sin(Ry);
    double cz=cos(Rz);
    double sz=sin(Rz);

    Mm d_r1_d_R = (BR(0.0), -cz * sy, -cy * sz);
    Mm d_r2_d_R = (BR(cx * cz * sy + sx * sz), cy * cz * sx, -cx * cz - sx * sy * sz);
    Mm d_r3_d_R = (BR(-cz * sx * sy + cx * sz),  cx * cy * cz, cz * sx - cx * sy * sz);
    Mm d_r4_d_R = (BR(0.0), -sy * sz, cy * cz);
    Mm d_r5_d_R = (BR(-cz * sx + cx * sy * sz), cy * sx * sz, cz * sx * sy - cx * sz);
    Mm d_r6_d_R = (BR(-cx * cz - sx * sy * sz), cx * cy * sz, cx * cz * sy + sx *sz);
    Mm d_r7_d_R = (BR(0.0), -cy, 0.0);
    Mm d_r8_d_R = (BR(cx * cy), -sx * sy, 0.0);
    Mm d_r9_d_R = (BR(-cy * sx), -cx * sy, 0.0);

    Mm trXYZworld=ctranspose(XYZworld);
    Mm d_Xc_d_R = trXYZworld * (BR(d_r1_d_R),semi,d_r2_d_R,semi,d_r3_d_R);
    Mm d_Yc_d_R = trXYZworld * (BR(d_r4_d_R),semi,d_r5_d_R,semi,d_r6_d_R);
    Mm d_Zc_d_R = trXYZworld * (BR(d_r7_d_R),semi,d_r8_d_R,semi,d_r9_d_R);
    
    H(colon(1,1,1),colon(col,1,col+2))=xfactor*ctranspose((d_Xc_d_R - Xu*d_Zc_d_R)*Zc_inv);
    H(colon(2,1,2),colon(col,1,col+2))=yfactor*ctranspose((d_Yc_d_R - Yu*d_Zc_d_R)*Zc_inv);

    MY_DEBUG(cout<<"Jacobian wrt to camera rotation angles:\n");
    MY_DEBUG(disp(H(colon(1,1,2),colon(col,1,col+2))));
    col+=3;
  }

  //wrt to focal length
  if (flag & D_FOCAL) {
    H(1,col) = Xu * Asp / dpx;
    H(2,col) = Yu / dpy;
    col++;
  }

  if (flag & D_CENTER) {
    H(1,colon(col,col+1)) = (BR(1.0), 0.0);
    H(2,colon(col,col+1)) = (BR(0.0), 1.0);
    col+=2;
  }

  if (flag & D_ASPECT) {
    H(1,col) = Xu * f / dpx;
    H(2,col) = 0;
    col++;
  }
  
}

/**************************************************************************
									   
  Starting from here, a series of Jacobian computation functions.

  The input to these function is a 3 vector given the (undistorted )image coord.
  (Xi,Yi) and the Z coordinate of the 3d point in Camera frame Zc (inverse of
  Zc, really.)

  For computation
  of the Jacobian wrt delta_angle (either of the camera or object),
  additional intermediate results, such as a point Po in {O} frame rotated by
  R_obj2world, or a point Pw in {W} frame rotated by R_world2cam, are
  also given as input to the function.
     
******************************************************************************/
void CAM::J_pos(Mm Xi_Yi_Zc /* 3x1 or 1x3 */,
	   Mw H /* 2x3 */)
{
  double Xi=Xi_Yi_Zc.r(1);
  double Yi=Xi_Yi_Zc.r(2);
  double Zc_inv=Xi_Yi_Zc.r(3);

  double xfactor = f * Asp / dpx;
  double yfactor = f / dpy ;
  
  Mm rot_row1=(BR(r1),r2,r3);
  Mm rot_row2=(BR(r4),r5,r6);
  Mm rot_row3=(BR(r7),r8,r9);

  // d_xi_d_Pw
  H(1,colon(1,1,3)) = Zc_inv*(xfactor * rot_row1 - (Xi-Cx) * rot_row3);
  // d_yi_d_Pw
  H(2,colon(1,1,3)) = Zc_inv*(yfactor * rot_row2 - (Yi-Cy) * rot_row3);
  
}


void CAM::J_da(Mm Xi_Yi_Zc /* 3x1 or 1x3 */,
	  Mw H /* 2x3 */,
	  Mm R_obj2world_Po /* 1x3 or 3x1 */)
{
  Mm H_pos;

  J_pos(Xi_Yi_Zc,Mw(H_pos));
  J_da_given_J_pos(H_pos, H, R_obj2world_Po);
  
}


void CAM::J_cam_pos(Mm Xi_Yi_Zc /* 3x1 or 1x3 */,
	       Mw H /* 2x3 */)
{
  double Xi=Xi_Yi_Zc.r(1);
  double Yi=Xi_Yi_Zc.r(2);
  double Zc_inv=Xi_Yi_Zc.r(3);

  double xfactor = f * Asp / dpx;
  double yfactor = f / dpy ;

  H(1,colon(1,1,3))= (BR(xfactor * Zc_inv), 0.0, -(Xi-Cx)*Zc_inv);
  H(2,colon(1,1,3))= (BR(0.0), yfactor * Zc_inv, -(Yi-Cy)*Zc_inv);
  
}

void CAM::J_cam_delta_angle(Mm Xi_Yi_Zc /* 3x1 or 1x3 */,
		       Mw H /* 2x3 */,
		       Mm R_world2cam_Pw /* 1x3 or 3x1 */)
{
  Mm H_cam_pos;

  J_cam_pos(Xi_Yi_Zc, Mw(H_cam_pos));
  J_cam_delta_angle_given_J_cam_pos(H_cam_pos, H, R_world2cam_Pw);
  
}
#endif
