#ifndef _CAM_H_
#define _CAM_H_
//
// File:      cam.h
// Author:    Xing Cindy Chen <xcchen@graphics.stanford.edu>
//
// Description:  routines related camera geometry
//

#include <math.h>
#include <string.h>
#include <sgl/vecmath3.h>
#include <sgl/transform3.h>
#ifdef USEMATLIB
#include "mymatlib.h"
#endif
#include "Quaternion.hh"

const char D_POINT = 0x01;
const char D_POS = 0x02;
const char D_ROT = 0x04;
const char D_FOCAL = 0x08;
const char D_CENTER = 0x10;
const char D_ASPECT = 0x20;

//==========================================================================

#define near near_xxx
#define far far_xxx

class CAM
{
public:
  char name[512];

  double Sx; //    effective CCD chip size in horizontal direction [mm]
  double Sy; //    effective CCD chip size in vertical direction [mm]

  double    dpx;              /* [mm/pix]  Effective X dimension of pixel in
				 frame grabber   */
  double    dpy;              /* [mm/pix]  Effective Y dimension of pixel in
				 frame grabber   */
  double    Cx;               /* [pix]     Z axis intercept of camera coordinate system      */
  double    Cy;               /* [pix]     Z axis intercept of camera coordinate system      */
  double    Asp;               /* []  Aspect ratio,  Scale factor to compensate for any error in dpx  */

  double    f;                /* [mm]          */

  double    Tx;               /* [mm]          */
  double    Ty;               /* [mm]          */
  double    Tz;               /* [mm]          */
  double    Rx;               /* [rad]         */
  double    Ry;               /* [rad]         */
  double    Rz;               /* [rad]         */

  Quaternion Rquat;  /* rotation represented by quaternion instead of Euler angles Rx,Ry,Rz */

  double rotm[3][3]; //entries of rotation matrix
 
  #define r1 rotm[0][0]
  #define r2 rotm[0][1]
  #define r3 rotm[0][2]
  #define r4 rotm[1][0]
  #define r5 rotm[1][1]
  #define r6 rotm[1][2]
  #define r7 rotm[2][0]
  #define r8 rotm[2][1]
  #define r9 rotm[2][2]
  

  /**********************************************************************/
  /* Class Transform3 is used to represent the 3x4 projection matrix P
     P.T is a 3x3 matrix representing the first 3 columns of P and 
     P.tranlation is a 3x1 vector representing the last column. 
     invP is a 3x3 matrix that presents the inverse of P.T.
     */
  Transform3 P; //3x4 world to image transformation matrix
  Matrix3  invP; // inverse of 3x3 part of P
  Point3 C; // 3x1 vector
  
  /********************** METHODS ***************************************/ 
 
public:
  
  CAM() {}
  virtual ~CAM() {};

  virtual void writecamfile(char *fname,bool use_quaternion=false) = 0;
  
  void error(char *s,char *s2 = "");
  
  char *getName() {return name;}
  void setName(char *name_) { sprintf(name,"%s",name_); }
  
  void getCamCenter(double *camCenter) const; // return 3x1 vector
  Point3 getCamCenter() const {return C;}
  void getCamSelfEulerAngle(double *angle) const; // return 3x1 vector
  void getOpAxis(double *opaxis); //return 3x1 (normalized) vector
  void getUpVector(double *up);
  
  double getFOVx(double dummy=0) const // FOV in degrees
    { return(atan2(Sx*0.5/Asp,f)*2.0/M_PI*180); }
  double getFOVy(double dummy=0) const
    { return(atan2(Sy*0.5,f)*2.0/M_PI*180); }
  void getProjMatrix(double *Proj) const;//3x4  column-major matrix version of P

  /* The following two functions are needed for setting up GL viewing
     parameters to simulate the view from a camera. Given the image
     size to be (imXsize,imYsize), and the near clipping plane NEAR, do
     the following in GL:

     glViewport(0,0,imXsize,imYsize);
     
     glMatrixMode(GL_PROJECTION);
     glLoadIdentity();	
     double near=NEAR;
     double far=FAR;
     double left,right,bottom,top;
     mycam.getGLFrustumParams(imXsize,imYsize,near,left,right,bottom,top);
     glFrustum(left,right,bottom,top,near,far);
     glRotated(180,1,0,0); //Don't forget!!! Camera y and world y are inverse.
     
     glMatrixMode(GL_MODELVIEW);
     glLoadIdentity();
     double modelview[16];
     camera.getGLModelViewMatrix(modelview);
     glMultMatrixd(modelview);

  */
  void getGLModelViewMatrix(double *modelView) const;// 4x4 column-major matrix
  
  void getGLFrustumParams(int imXsize, int imYsize, double near,
			  double &left,double &right,double &bottom,double &top) const ;

  void getIntrinsicMatrix(double A[3][3]) const;
  
  void setPos(double Tx_, double Ty_, double Tz_);
  void setCamCenter(double cx,double cy , double cz) {
    //NORE:: This depends on rotation matrix R. so if one needs to set
    //both rotation and translation, one need to set rotation first!
    //since C=-R'*T, therefore T=-R*C
    double tx=-(r1*cx+r2*cy+r3*cz);
    double ty=-(r4*cx+r5*cy+r6*cz);
    double tz=-(r7*cx+r8*cy+r9*cz);
    setPos(tx,ty,tz);
  }
  void setCamSelfEulerAngle(double ax,double ay,double az) {
    //NOTE: since we change the camera itself rather than the world,
    //the coordinates of world center in camera frame (Tx,Ty,Tz)
    // has changed ALSO,
    // but camera center in world frame (C) shouldn't change.
    double c[3];
    getCamCenter(c);
    setRot(-ax,-ay,-az);
    setCamCenter(c[0],c[1],c[2]);
  }
  
  void setFOVx(double degrees,double xsize_in_pixel=0/*dummy*/)
    { f = (Sx*0.5/Asp)/tan(0.5*degrees*M_PI/180.0); }
  void setFOVy(double degrees,double ysize_in_pixel=0/*dummy*/)
    { f = (Sy*0.5)/tan(0.5*degrees*M_PI/180.0); }
  void setRot(double Rx_, double Ry_, double Rz_);
  void setRotQuat(Quaternion quat);
  void setRotQuat(double w,double x,double y,double z)
    {setRotQuat(Quaternion(w,x,y,z)); }

  /* set BOTH the rotation/translation so that camera  is located
     at (xfrom,yfrom,zfrom) and pointing at (xto,yto,zto) */
  void lookAt(double xfrom,double yfrom,double zfrom,
		 double xto,double yto,double zto);
   void lookAt(double xfrom,double yfrom,double zfrom,
		 double xto,double yto,double zto,
	       double xup,double yup,double zup); 
  
  /* adjust the pos/orientation so that they are relative to a reference camera:
     
     RotMatrixNew = RotMatrixOld * RotMatrixOfRefCamera;
     TranslationNew = RotationMatrixOld * TranslationOfRefCamera + TranslationOld
  */
  void adjustToRefCamera(CAM& refcam);
    
  
  virtual void distorted_to_undistorted_sensor_coord(double, double, 
						     double &, double &){}
  virtual void undistorted_to_distorted_sensor_coord(double, double, 
						     double &, double &){}
  

  void undistorted_to_distorted_image_coord (double Xfu, double Yfu, 
					     double &Xfd, double &Yfd);
  void distorted_to_undistorted_image_coord (double Xfd, double Yfd, 
					     double &Xfu, double &Yfu);

  void imageToWorld(double ximage, double yimage, double zworld,
		    double &xworld, double &yworld);
  void imageToWorld(double ximage, double yimage, 
		    double &xworld, double &yworld);
  
  void worldToCamera(double xworld, double yworld, double zworld,
		     double& xc, double& yc,double& zc);

  /* project world point to image, return Z coordinates in camera frame
     so that one can decide whether the point is in front of the camera
     (Zc>0) or  behind the camera (Zc<0) */
  double worldToImage(double xworld, double yworld, double zworld,
		    double &ximage, double &yimage);
  double worldToUndistortedImage(double xworld, double yworld, double zworld,
			       double &ximage, double &yimage);


  /* given a 2d point on the image, returns 3d line that connects
     the camera center and the 2d point, in terms of the starting point
     (i.e. camera center) and the direction. */
  void imageToLine3d(double ximage,double yimage,
		     double *startPoint,double *direction);
  void undistortedImageToLine3d(double ximage,double yimage,
		     double *startPoint,double *direction);
  
  /* project the 3D line defined by the center of camera cam1 and a 2D
     image point (ximage1,yimage1) seen by cam1 to a 2D line on
     the image plane of "this" camera */
  void epipolarLine(CAM &cam1, 
		    double ximage1, double yimage1,
		    double &xbegin, double &ybegin,
		    double &xend, double &yend);

  int point3d(double ximage0,double yimage0,
	CAM &cam1, double ximage1, double yimage1,
	double &x3d,double &y3d, double &z3d,double &err);

#ifdef USEMATLIB
  void setPos(Mm pos) { setPos(pos.r(1),pos.r(2),pos.r(3)); }
  void setRot(Mm rot) { setRot(rot.r(1),rot.r(2),rot.r(3)); }
  double worldToUndistortedImage(Mm XYZworld /* 3x1 */,
				Mw Xi_Yi_Zc /* 1x3 */,
				Mw R_XYZworld /* 1x3 */);
  
  /* Given world coord. of a point, compute undistorted image coord.
  and its jacobian over point position, camera parameters according
  to the bits set in the flag */
  void undistoredImageJacobian(Mm XYZworld /* 3x1 vector */,
			double& ximage, double& yimage,
			Mm& H, char flag);
 
  /* start from here, a series of Jacobian computation functions,
     each is given the projected point Xi,Yi and intermediate result Zc,
     it is also given intermediate results such as rotated object
     cordinates and rotated world coordinates. */
  void J_pos(Mm Xi_Yi_Zc /* 3x1 or 1x3 */,
	     Mw H /* 2x3 */);
  void J_da(Mm Xi_Yi_Zc /* 3x1 or 1x3 */, 
		Mw H /* 2x3 */, 
		Mm R_obj2world_Po /* 1x3 or 3x1 */); 

  void J_da_given_J_pos(Mm H_pos /* 2x3 Jacobian wrt obj pos */,
	  Mw H /* 2x3 Jacobian wrt obj delta_angle */,
	  Mm R_obj2world_Po) {
	H(1,colon(1,1,3))=mycross(R_obj2world_Po, H_pos(1,colon(1,1,3)));
	H(2,colon(1,1,3))=mycross(R_obj2world_Po, H_pos(2,colon(1,1,3))); }

  void J_cam_pos(Mm Xi_Yi_Zc /* 3x1 or 1x3 */,
		Mw H /* 2x3 */);
  void J_cam_delta_angle(Mm Xi_Yi_Zc /* 3x1 or 1x3 */, 
		Mw H /* 2x3 */, 
		Mm R_world2cam_Pw /* 1x3 or 3x1 */); 

  void J_cam_delta_angle_given_J_cam_pos(Mm H_cam_pos /* 2x3 J. wrt cam_pos */,
		Mw H /* 2x3 */, 
		Mm R_world2cam_Pw /* 1x3 or 3x1 */) {
	H(1,colon(1,1,3))=mycross(R_world2cam_Pw, H_cam_pos(1,colon(1,1,3)));
	H(2,colon(1,1,3))=mycross(R_world2cam_Pw, H_cam_pos(2,colon(1,1,3)));}
#endif
  
protected:
  void updateProjMatrixAndCamCenter() {
       	  Matrix3 rotation(r1,r2,r3,
			   r4, r5,r6,
			   r7,r8,r9);
	  Vector3 translate(Tx,Ty,Tz);
	  Matrix3 intrinsic(f*Asp/dpx, 0.0,   Cx,
			    0.0,       f/dpy,  Cy, 
			    0.0,       0.0,  1.0);
	  //NOTE::A*B is actually BA for two matrices!!!
	  P=Transform3(rotation*intrinsic, intrinsic.apply(translate));
	  invP = P.Matrix3::inverse(); //just invert the 3x3 matrix of P 	
	  C=invP.apply((Point3)(-P.translation));
	  //C=-invP*P(colon(1,1,3),4);
	  //C=-ctranspose(rotation)*translate;
  } 
  void rotMatrix2Euler(double r[3][3],double& anglex,double& angley,double& anglez);
};


class CAM_Tsai : public CAM 
{

protected:
  double    kappa1;           /* [1/mm^2]      */

  double    Ncx;              /* [sel]     Number of sensor elements in camera's x direction */
  double    Nfx;              /* [pix]     Number of pixels in frame grabber's x direction   */
  double    dx;               /* [mm/sel]  X dimension of camera's sensor element (in mm)    */
  double    dy;               /* [mm/sel]  Y dimension of camera's sensor element (in mm)    */

public:
  CAM_Tsai(){}
  CAM_Tsai(char *fname);
  
  void writecamfile(char *fname,bool use_quaternion=false);
  
  void distorted_to_undistorted_sensor_coord (double Xd, double Yd, 
					      double &Xu, double &Yu);
  void undistorted_to_distorted_sensor_coord (double Xu, double Yu, 
					      double &Xd, double &Yd);  


};


class CAM_Finland: public CAM 
{

public:
  double NDX; //   number of pixels in horizontal direction
  double NDY; //   number of pixels in vertical direction

  double Rad1,Rad2,Tan1,Tan2; // lens distortion parameters
  double a1,a2,a3,a4,a5,a6,a7,a8; // inverse lens distortion parameters
public:
  CAM_Finland();
  CAM_Finland(CAM_Finland &cam);
  CAM_Finland(char *fname);
  
  void writecamfile(char *fname, bool use_quaternion = false);
     
  void distorted_to_undistorted_sensor_coord (double Xd, double Yd, 
					      double &Xu, double &Yu);
  void undistorted_to_distorted_sensor_coord (double Xu, double Yu, 
					      double &Xd, double &Yd);    
};




#endif /* _CAM_H_ */
