//
// transform3.cc
//
// $Id: transform3.cc,v 1.1.1.1 2001/02/28 00:28:40 cstolte Exp $
//

#include <sgl/transform3.h>
#include <sgl/mathfunc.h>

template <class T>
inline void swap(T& a, T& b) {
    T tmp = a;
    a = b;
    b = tmp;
}

const Matrix3 Matrix3::unit(1, 0, 0, 0, 1, 0, 0, 0, 1);
const Transform3 Transform3::unit(Matrix3::unit);

//
// Matrix3
//

Matrix3::Matrix3(double a00, double a01, double a02,
		 double a10, double a11, double a12,
		 double a20, double a21, double a22) {
  T[0][0] = a00;  T[0][1] = a01;  T[0][2] = a02;
  T[1][0] = a10;  T[1][1] = a11;  T[1][2] = a12;
  T[2][0] = a20;  T[2][1] = a21;  T[2][2] = a22;
}

std::ostream &
operator<<(std::ostream &os, const Matrix3 &m) {
    os << "[[" << m.T[0][0] << ' ' << m.T[0][1] << ' ' << m.T[0][2] << "]" << std::endl;
    os << " [" << m.T[1][0] << ' ' << m.T[1][1] << ' ' << m.T[1][2] << "]" << std::endl;
    os << " [" << m.T[2][0] << ' ' << m.T[2][1] << ' ' << m.T[2][2] << "]]" << std::endl;
    return os;
}

Matrix3
Matrix3::operator*(const Matrix3 &m) const {
    Matrix3 ret(*this);
    ret *= m;
    return ret;
}

Matrix3
Matrix3::operator+(const Matrix3 &m) const {
    Matrix3 ret(*this);
    ret += m;
    return ret;
}

Matrix3
Matrix3::operator-(const Matrix3 &m) const {
    Matrix3 ret(*this);
    ret -= m;
    return ret;
}

Matrix3 &
Matrix3::operator*=(const Matrix3 &m) {
    double tmp[3][3];
    tmp[0][0] = m.T[0][0] * T[0][0] + m.T[0][1] * T[1][0] + m.T[0][2] * T[2][0];
    tmp[0][1] = m.T[0][0] * T[0][1] + m.T[0][1] * T[1][1] + m.T[0][2] * T[2][1];
    tmp[0][2] = m.T[0][0] * T[0][2] + m.T[0][1] * T[1][2] + m.T[0][2] * T[2][2];
    tmp[1][0] = m.T[1][0] * T[0][0] + m.T[1][1] * T[1][0] + m.T[1][2] * T[2][0];
    tmp[1][1] = m.T[1][0] * T[0][1] + m.T[1][1] * T[1][1] + m.T[1][2] * T[2][1];
    tmp[1][2] = m.T[1][0] * T[0][2] + m.T[1][1] * T[1][2] + m.T[1][2] * T[2][2];
    tmp[2][0] = m.T[2][0] * T[0][0] + m.T[2][1] * T[1][0] + m.T[2][2] * T[2][0];
    tmp[2][1] = m.T[2][0] * T[0][1] + m.T[2][1] * T[1][1] + m.T[2][2] * T[2][1];
    tmp[2][2] = m.T[2][0] * T[0][2] + m.T[2][1] * T[1][2] + m.T[2][2] * T[2][2];

    for (int i = 0; i < 3; ++i)
	for (int j = 0; j < 3; ++j)
	    T[i][j] = tmp[i][j];
    return *this;
}

Matrix3 &
Matrix3::operator+=(const Matrix3 &m) {
    T[0][0] += m.T[0][0];
    T[0][1] += m.T[0][1];
    T[0][2] += m.T[0][2];
    T[1][0] += m.T[1][0];
    T[1][1] += m.T[1][1];
    T[1][2] += m.T[1][2];
    T[2][0] += m.T[2][0];
    T[2][1] += m.T[2][1];
    T[2][2] += m.T[2][2];
    return *this;
}

Matrix3 &
Matrix3::operator-=(const Matrix3 &m) {
    T[0][0] -= m.T[0][0];
    T[0][1] -= m.T[0][1];
    T[0][2] -= m.T[0][2];
    T[1][0] -= m.T[1][0];
    T[1][1] -= m.T[1][1];
    T[1][2] -= m.T[1][2];
    T[2][0] -= m.T[2][0];
    T[2][1] -= m.T[2][1];
    T[2][2] -= m.T[2][2];
    return *this;
}

Matrix3 &
Matrix3::identity() {
    T[0][0] = 1;  T[0][1] = 0;  T[0][2] = 0;
    T[1][0] = 0;  T[1][1] = 1;  T[1][2] = 0;
    T[2][0] = 0;  T[2][1] = 0;  T[2][2] = 1;
    return *this;
}

Matrix3 &
Matrix3::zero() {
    T[0][0] = 0;  T[0][1] = 0;  T[0][2] = 0;
    T[1][0] = 0;  T[1][1] = 0;  T[1][2] = 0;
    T[2][0] = 0;  T[2][1] = 0;  T[2][2] = 0;
    return *this;
}

Matrix3 &
Matrix3::rotatex(double theta) {
    return *this *= ::rotatex(theta);
}

Matrix3 &
Matrix3::rotatey(double theta) {
    return *this *= ::rotatey(theta);
}

Matrix3 &
Matrix3::rotatez(double theta) {
    return *this *= ::rotatez(theta);
}

Matrix3 &
Matrix3::rotate(double theta, const Vector3 &a) {
    return *this *= ::rotate(theta, a);
}

Matrix3 &
Matrix3::scale(double sx, double sy, double sz) {
    return *this *= ::scale(sx, sy, sz);
}

Matrix3 &
Matrix3::transpose() {
    swap(T[0][1], T[1][0]);
    swap(T[0][2], T[2][0]);
    swap(T[1][2], T[2][1]);
    return *this;
}

Matrix3 &
Matrix3::invert() {
    double tmp[3][3];
    double d = determinant();

    tmp[0][0] = (T[1][1] * T[2][2] - T[1][2] * T[2][1]) / d;
    tmp[0][1] = (T[1][0] * T[2][2] - T[1][2] * T[2][0]) / -d;
    tmp[0][2] = (T[1][0] * T[2][1] - T[1][1] * T[2][0]) / d;

    tmp[1][0] = (T[0][1] * T[2][2] - T[0][2] * T[2][1]) / -d;
    tmp[1][1] = (T[0][0] * T[2][2] - T[0][2] * T[2][0]) / d;
    tmp[1][2] = (T[0][0] * T[2][1] - T[0][1] * T[2][0]) / -d;

    tmp[2][0] = (T[0][1] * T[1][2] - T[0][2] * T[1][1]) / d;
    tmp[2][1] = (T[0][0] * T[1][2] - T[0][2] * T[1][0]) / -d;
    tmp[2][2] = (T[0][0] * T[1][1] - T[0][1] * T[1][0]) / d;

    for (int i = 0; i < 3; ++i)
	for (int j = 0; j < 3; ++j)
	    // note that tmp should be transposed before assigning
	    // to *this, so just do it here...
	    T[i][j] = tmp[j][i];
    return *this;
}

double
Matrix3::determinant() const {
    return ((T[0][0] * (T[1][1] * T[2][2] - T[1][2] * T[2][1])) -
	    (T[0][1] * (T[1][0] * T[2][2] - T[1][2] * T[2][0])) +
	    (T[0][2] * (T[1][0] * T[2][1] - T[1][1] * T[2][0])));
}

double 
Matrix3::determinant(const Matrix3 &adj) const {
    return T[0][0] * adj.T[0][0] 
    	+ T[0][1] * adj.T[1][0] 
	+ T[0][2] * adj.T[2][0];
}

Matrix3 
Matrix3::adjoint() const {
    Matrix3 ret;
    double sign = 1;
    for (u_int i = 0; i < 3; ++i) {
	for (u_int j = 0; j < 3; ++j) {
	    ret.T[i][j] = cofactor(i, j) * sign; // save the transpose
	    sign = -sign;
	}
    }
    return ret;
}

Point3
Matrix3::apply(const Point3 &p) const {
    return Point3(T[0][0] * p.x + T[0][1] * p.y + T[0][2] * p.z,
		  T[1][0] * p.x + T[1][1] * p.y + T[1][2] * p.z,
		  T[2][0] * p.x + T[2][1] * p.y + T[2][2] * p.z);
}

Vector3
Matrix3::apply(const Vector3 &v) const {
    return Vector3(T[0][0] * v.x + T[0][1] * v.y + T[0][2] * v.z,
		   T[1][0] * v.x + T[1][1] * v.y + T[1][2] * v.z,
		   T[2][0] * v.x + T[2][1] * v.y + T[2][2] * v.z);
}

Normal3
Matrix3::apply(const Normal3 &n, const Matrix3 *inv) const {
    if (inv) {
	return Normal3(n.x * inv->T[0][0] + n.y * inv->T[1][0] + n.z * inv->T[2][0],
		       n.x * inv->T[0][1] + n.y * inv->T[1][1] + n.z * inv->T[2][1],
		       n.x * inv->T[0][2] + n.y * inv->T[1][2] + n.z * inv->T[2][2]);
    }
    else {
	Matrix3 inv1 = inverse();
	// skip transpose of inverse and take dot with columns
	return Normal3(n.x * inv1.T[0][0] + n.y * inv1.T[1][0] + n.z * inv1.T[2][0],
		       n.x * inv1.T[0][1] + n.y * inv1.T[1][1] + n.z * inv1.T[2][1],
		       n.x * inv1.T[0][2] + n.y * inv1.T[1][2] + n.z * inv1.T[2][2]);
    }
}

bool
Matrix3::operator==(const Matrix3 &m2) const {
    return (T[0][0] == m2.T[0][0] &&
	    T[0][1] == m2.T[0][1] &&
	    T[0][2] == m2.T[0][2] &&
	    T[1][0] == m2.T[1][0] &&
	    T[1][1] == m2.T[1][1] &&
	    T[1][2] == m2.T[1][2] &&
	    T[2][0] == m2.T[2][0] &&
	    T[2][1] == m2.T[2][1] &&
	    T[2][2] == m2.T[2][2]);
}

bool
Matrix3::operator!=(const Matrix3 &m2) const {
    return (T[0][0] != m2.T[0][0] ||
	    T[0][1] != m2.T[0][1] ||
	    T[0][2] != m2.T[0][2] ||
	    T[1][0] != m2.T[1][0] ||
	    T[1][1] != m2.T[1][1] ||
	    T[1][2] != m2.T[1][2] ||
	    T[2][0] != m2.T[2][0] ||
	    T[2][1] != m2.T[2][1] ||
	    T[2][2] != m2.T[2][2]);
}

double
Matrix3::cofactor(int i, int j) const {
    assert(i >= 0 && i < 3);
    assert(j >= 0 && j < 3);

    switch (i) {
      case 0:
	switch (j) {
	  case 0:
	    return T[1][1] * T[2][2] - T[1][2] * T[2][1];
	  case 1:
	    return T[1][0] * T[2][2] - T[1][2] * T[2][0];
	  case 2:
	    return T[1][0] * T[2][1] - T[1][1] * T[2][0];
	}
      case 1:
	switch (j) {
	  case 0:
	    return T[0][1] * T[2][2] - T[0][2] * T[2][1];
	  case 1:
	    return T[0][0] * T[2][2] - T[0][2] * T[2][0];
	  case 2:
	    return T[0][0] * T[2][1] - T[0][1] * T[2][0];
	}
      case 2:
	switch(j) {
	  case 0:
	    return T[0][1] * T[1][2] - T[0][2] * T[1][1];
	  case 1:
	    return T[0][0] * T[1][2] - T[0][2] * T[1][0];
	  case 2:
	    return T[0][0] * T[1][1] - T[0][1] * T[1][0];
	}
    }
    return 0; // notreached
}

Matrix3 identityM3() {
    return Matrix3(1, 0, 0,
		   0, 1, 0,
		   0, 0, 1);
}

Matrix3 rotatex(double theta) {
    double s = sin(theta);
    double c = cos(theta);
    return Matrix3(1, 0, 0,
		   0, c, -s,
		   0, s, c);
}

Matrix3 rotatey(double theta) {
    double s = sin(theta);
    double c = cos(theta);
    return Matrix3(c, 0, s,
		   0, 1, 0,
		  -s, 0, c);
}

Matrix3 rotatez(double theta) {
    double s = sin(theta);
    double c = cos(theta);
    return Matrix3(c, -s, 0,
		   s,  c, 0,
		   0,  0, 1);
}

Matrix3 rotate(double theta, const Vector3 &a) {
    Vector3 axis = a.hat();
    double s = sin(theta);
    double c = cos(theta);

    return Matrix3((axis.x * axis.x + (1. - axis.x * axis.x) * c),
		   (axis.x * axis.y * (1 - c) - axis.z * s),
		   (axis.x * axis.z * (1 - c) + axis.y * s),

		   (axis.x * axis.y * (1 - c) + axis.z * s),
		   (axis.y * axis.y + (1 - axis.y * axis.y) * c),
		   (axis.y * axis.z * (1 - c) - axis.x * s),

		   (axis.x * axis.z * (1 - c) - axis.y * s),
		   (axis.y * axis.z * (1 - c) + axis.x * s),
		   (axis.z * axis.z + (1 - axis.z * axis.z) * c));
}

Matrix3 scale(double sx, double sy, double sz) {
    return Matrix3(sx, 0,  0,
		   0, sy,  0,
		   0,  0, sz);
}

//
// Transform3
//

Transform3::Transform3(const Matrix3 &m) 
  : Matrix3(m), translation(0, 0, 0) {
}

Transform3::Transform3(const Matrix3 &m, const Vector3 &t) 
  : Matrix3(m), translation(t) {
}

std::ostream &
operator<<(std::ostream &os, const Transform3 &tr) {
    os << (const Matrix3 &)tr << std::endl << tr.translation;
    return os;
}

Transform3
Transform3::operator*(const Transform3 &t) const {
    Transform3 ret(*this);
    ret *= t;
    return ret;
}

Transform3 &
Transform3::operator*=(const Transform3 &t) {
    double tmp[3][3];
    tmp[0][0] = t.T[0][0] * T[0][0] + t.T[0][1] * T[1][0] + t.T[0][2] * T[2][0];
    tmp[0][1] = t.T[0][0] * T[0][1] + t.T[0][1] * T[1][1] + t.T[0][2] * T[2][1];
    tmp[0][2] = t.T[0][0] * T[0][2] + t.T[0][1] * T[1][2] + t.T[0][2] * T[2][2];
    tmp[1][0] = t.T[1][0] * T[0][0] + t.T[1][1] * T[1][0] + t.T[1][2] * T[2][0];
    tmp[1][1] = t.T[1][0] * T[0][1] + t.T[1][1] * T[1][1] + t.T[1][2] * T[2][1];
    tmp[1][2] = t.T[1][0] * T[0][2] + t.T[1][1] * T[1][2] + t.T[1][2] * T[2][2];
    tmp[2][0] = t.T[2][0] * T[0][0] + t.T[2][1] * T[1][0] + t.T[2][2] * T[2][0];
    tmp[2][1] = t.T[2][0] * T[0][1] + t.T[2][1] * T[1][1] + t.T[2][2] * T[2][1];
    tmp[2][2] = t.T[2][0] * T[0][2] + t.T[2][1] * T[1][2] + t.T[2][2] * T[2][2];

    for (int i = 0; i < 3; ++i)
	for (int j = 0; j < 3; ++j)
	    T[i][j] = tmp[i][j];

    double tx = t.translation.x +
	translation.x * t.T[0][0] + translation.y * t.T[0][1] +
	    translation.z * t.T[0][2];
    double ty = t.translation.y +
	translation.x * t.T[1][0] + translation.y * t.T[1][1] +
	    translation.z * t.T[1][2];
    double tz = t.translation.z +
	translation.x * t.T[2][0] + translation.y * t.T[2][1] +
	    translation.z * t.T[2][2];
    translation.x = tx;
    translation.y = ty;
    translation.z = tz;

    return *this;
}

Transform3
operator*(const Matrix3 &mat, const Transform3 &tr)
{
    Transform3 ret(mat);
    return ret *= tr;
}

Transform3 
Transform3::operator*(const Matrix3 &m) const {
    Transform3 ret(*this);
    return ret *= m;
}

Transform3 &
Transform3::operator*=(const Matrix3 &m) {
    double tmp[3][3];
    tmp[0][0] = m.T[0][0] * T[0][0] + m.T[0][1] * T[1][0] + m.T[0][2] * T[2][0];
    tmp[0][1] = m.T[0][0] * T[0][1] + m.T[0][1] * T[1][1] + m.T[0][2] * T[2][1];
    tmp[0][2] = m.T[0][0] * T[0][2] + m.T[0][1] * T[1][2] + m.T[0][2] * T[2][2];
    tmp[1][0] = m.T[1][0] * T[0][0] + m.T[1][1] * T[1][0] + m.T[1][2] * T[2][0];
    tmp[1][1] = m.T[1][0] * T[0][1] + m.T[1][1] * T[1][1] + m.T[1][2] * T[2][1];
    tmp[1][2] = m.T[1][0] * T[0][2] + m.T[1][1] * T[1][2] + m.T[1][2] * T[2][2];
    tmp[2][0] = m.T[2][0] * T[0][0] + m.T[2][1] * T[1][0] + m.T[2][2] * T[2][0];
    tmp[2][1] = m.T[2][0] * T[0][1] + m.T[2][1] * T[1][1] + m.T[2][2] * T[2][1];
    tmp[2][2] = m.T[2][0] * T[0][2] + m.T[2][1] * T[1][2] + m.T[2][2] * T[2][2];
    translation = 
	Vector3(translation.x * m.T[0][0] + translation.y * m.T[0][1] +
		translation.z * m.T[0][2],
		translation.x * m.T[1][0] + translation.y * m.T[1][1] +
		translation.z * m.T[1][2],
		translation.x * m.T[2][0] + translation.y * m.T[2][1] +
		translation.z * m.T[2][2]);

    for (int i = 0; i < 3; ++i)
	for (int j = 0; j < 3; ++j)
	    T[i][j] = tmp[i][j];
    return *this;
}

Transform3 &
Transform3::identity() {
    translation = Vector3(0,0,0);
    Matrix3::identity();
    return *this;
}

Transform3 &
Transform3::translate(double tx, double ty, double tz) {
    return *this *= ::translate(tx, ty, tz);
}

Transform3 &
Transform3::rotatex(double angle) {
    return *this *= ::rotatex(angle);
}

Transform3 &
Transform3::rotatey(double angle) {
    return *this *= ::rotatey(angle);
}

Transform3 &
Transform3::rotatez(double angle) {
    return *this *= ::rotatez(angle);
}

Transform3 &
Transform3::rotate(double theta, const Vector3 &a) {
    return *this *= ::rotate(theta, a);
}

Transform3 &
Transform3::scale(double sx, double sy, double sz) {
    return *this *= ::scale(sx, sy, sz);
}

Transform3 &
Transform3::invert() {
    Matrix3::invert();
    translation = Vector3(-(translation.x * T[0][0] + translation.y * T[0][1] +
			    translation.z * T[0][2]),
			  -(translation.x * T[1][0] + translation.y * T[1][1] +
			    translation.z * T[1][2]),
			  -(translation.x * T[2][0] + translation.y * T[2][1] +
			    translation.z * T[2][2]));
    return *this;
}

Transform3&
Transform3::camera(const Point3 &at, const Point3 &from,
		   const Point3 &up) {
    // apply rotation and translation to convert to a coordinate system
    // with the from point at the origin, the at point on the positive
    // z axis, and the up point on the positive y axis

    Vector3 Z = (at - from).normalize();
    Vector3 X = (cross(up - from, Z)).normalize();
    Vector3 Y = (cross(Z, X)).normalize();

    Matrix3 m(X.x, X.y, X.z,
	      Y.x, Y.y, Y.z,
	      Z.x, Z.y, Z.z);

    Vector3 v(-dot(from - Point3::origin, X),
	      -dot(from - Point3::origin, Y),
	      -dot(from - Point3::origin, Z));

    return *this *= Transform3(m, v);
}

Transform3&
Transform3::perspective(double fov, double aspect,
			double /*znear*/, double zfar) {
    // apply differential scaling to convert the above coordinate system
    // into the canonical perspective coordinate system for the given field
    // of view, aspect ratio, and near and far planes.
    // (Foley et al, chapter 6)

    double cotfov = tan(fov / 2.0);
    if (cotfov != 0.0)
	cotfov = 1.0 / cotfov;

    Matrix3 tmp(cotfov / aspect, 0.0,      0.0,
		0.0,             cotfov,   0.0,
		0.0,             0.0,      -1.0 / zfar);

    return *this *= tmp;
}

bool
Transform3::operator==(const Transform3 &t2) const {
    return (T[0][0] == t2.T[0][0] &&
	    T[0][1] == t2.T[0][1] &&
	    T[0][2] == t2.T[0][2] &&
	    T[1][0] == t2.T[1][0] &&
	    T[1][1] == t2.T[1][1] &&
	    T[1][2] == t2.T[1][2] &&
	    T[2][0] == t2.T[2][0] &&
	    T[2][1] == t2.T[2][1] &&
	    T[2][2] == t2.T[2][2] &&
	    translation == t2.translation);
}

bool
Transform3::operator!=(const Transform3 &t2) const {
    return (T[0][0] != t2.T[0][0] ||
	    T[0][1] != t2.T[0][1] ||
	    T[0][2] != t2.T[0][2] ||
	    T[1][0] != t2.T[1][0] ||
	    T[1][1] != t2.T[1][1] ||
	    T[1][2] != t2.T[1][2] ||
	    T[2][0] != t2.T[2][0] ||
	    T[2][1] != t2.T[2][1] ||
	    T[2][2] != t2.T[2][2] ||
	    translation != t2.translation);
}

Point3
Transform3::apply(const Point3 &p) const {
    Point3 ret(Matrix3::apply(p));
    return ret += translation;
}

Vector3
Transform3::apply(const Vector3 &v) const {
    return Matrix3::apply(v);
}

Normal3
Transform3::apply(const Normal3 &n, const Transform3 *inv) const {
    return Matrix3::apply(n, inv);
}

Ray3
Transform3::apply(const Ray3 &r) const {
    return Ray3(apply(r.o), apply(r.d));
}

Transform3 translate(double tx, double ty, double tz) {
    return Transform3(Matrix3::unit, Vector3(tx, ty, tz));
}

/*
 * Window Transform maps:
 *		l to -1
 *		r to +1
 *		b to -1
 *		t to +1
 */
//COTransform3&
//COTransform3::window(double l, double r, double b, double t) {
//CO    double sx,sy,dx,dy;
//CO    register double *aptr;
//CO    int register cnt;
//CO
//CO    sx=2./(r-l);
//CO    sy=2./(t-b);
//CO    dx=1.0 - r*sx;
//CO    dy=1.0 - t*sy;
//CO
//CO    aptr=T[0];
//CO    cnt =4;
//CO    do {
//CO        aptr[3] += aptr[0]*dx + aptr[1]*dy;
//CO        aptr[0] *= sx;
//CO        aptr[1] *= sy;
//CO        aptr += 4;
//CO    } while (--cnt);
//CO
//CO    return *this;
//CO}

//
// TransInv3
//

TransInv3::TransInv3(const Transform3 &t) 
  : trans(t) {
      itrans = trans.inverse();
}

TransInv3::TransInv3(const Transform3 &t, const Transform3 &i)
  : trans(t), itrans(i) {
}

Point3
TransInv3::apply(const Point3 &p) const {
    return trans.apply(p);
}

Vector3
TransInv3::apply(const Vector3 &v) const {
    return trans.apply(v);
}

Normal3
TransInv3::apply(const Normal3 &n) const {
    return trans.apply(n, &itrans);
}

Ray3
TransInv3::apply(const Ray3 &r) const {
    return trans.apply(r);
}


Point3
TransInv3::applyInv(const Point3 &p) const {
    return itrans.apply(p);
}

Vector3
TransInv3::applyInv(const Vector3 &v) const {
    return itrans.apply(v);
}

Normal3
TransInv3::applyInv(const Normal3 &n) const {
    return itrans.apply(n, &trans);
}

Ray3
TransInv3::applyInv(const Ray3 &r) const {
    return itrans.apply(r);
}

//
// Projection3
//

Projection3::Projection3(double a00, double a01, double a02, double a03,
			 double a10, double a11, double a12, double a13,
			 double a20, double a21, double a22, double a23,
			 double a30, double a31, double a32, double a33) {
    T[0][0] = a00; T[0][1] = a01; T[0][2] = a02; T[0][3] = a03;
    T[1][0] = a10; T[1][1] = a11; T[1][2] = a12; T[1][3] = a13;
    T[2][0] = a20; T[2][1] = a21; T[2][2] = a22; T[2][3] = a23;
    T[3][0] = a30; T[3][1] = a31; T[3][2] = a32; T[3][3] = a33;
}

Projection3::Projection3(const Matrix3 &m) {
    for (int i = 0; i < 3; ++i)
	for (int j = 0; j < 3; ++j)
	    T[i][j] = m.T[i][j];
    T[0][3] = T[1][3] = T[2][3] = T[3][0] = T[3][1] = T[3][2] = 0;
    T[3][3] = 1;
}

Projection3::Projection3(const Transform3 &t) {
    for (int i = 0; i < 3; ++i)
	for (int j = 0; j < 3; ++j)
	    T[i][j] = t.T[i][j];
    T[0][3] = t.translation.x;
    T[1][3] = t.translation.y;
    T[2][3] = t.translation.z;
    T[3][0] = T[3][1] = T[3][2] = 0;
    T[3][3] = 1;
}

Projection3
Projection3::operator*(const Projection3 &p) const {
    Projection3 ret(*this);
    return ret *= p;
}

Projection3
Projection3::operator+(const Projection3 &p) const {
    Projection3 ret(*this);
    return ret += p;
}

Projection3
Projection3::operator-(const Projection3 &p) const {
    Projection3 ret(*this);
    return ret -= p;
}

Projection3 &
Projection3::operator*=(const Projection3 &p2) {
    Projection3 old(*this);
    for (int i = 0; i < 4; ++i)
	for (int j = 0; j < 4; ++j) {
	    T[i][j] = 0;
	    for (int k = 0; k < 4; ++k) 
		T[i][j] += p2.T[i][k] * old.T[k][j]; 
	}
    return *this;
}

Projection3 &
Projection3::operator+=(const Projection3 &p) {
    T[0][0] += p.T[0][0]; T[0][1] += p.T[0][1]; T[0][2] += p.T[0][2]; T[0][3] += p.T[0][3];
    T[1][0] += p.T[1][0]; T[1][1] += p.T[1][1]; T[1][2] += p.T[1][2]; T[1][3] += p.T[1][3];
    T[2][0] += p.T[2][0]; T[2][1] += p.T[2][1]; T[2][2] += p.T[2][2]; T[2][3] += p.T[2][3];
    T[3][0] += p.T[3][0]; T[3][1] += p.T[3][1]; T[3][2] += p.T[3][2]; T[3][3] += p.T[3][3];
    return *this;
}

Projection3 & Projection3::operator-=(const Projection3 &p) {
    T[0][0] -= p.T[0][0]; T[0][1] -= p.T[0][1]; T[0][2] -= p.T[0][2]; T[0][3] -= p.T[0][3];
    T[1][0] -= p.T[1][0]; T[1][1] -= p.T[1][1]; T[1][2] -= p.T[1][2]; T[1][3] -= p.T[1][3];
    T[2][0] -= p.T[2][0]; T[2][1] -= p.T[2][1]; T[2][2] -= p.T[2][2]; T[2][3] -= p.T[2][3];
    T[3][0] -= p.T[3][0]; T[3][1] -= p.T[3][1]; T[3][2] -= p.T[3][2]; T[3][3] -= p.T[3][3];
    return *this;
}

Projection3 &
Projection3::identity() {
    T[0][0] = 1; T[0][1] = 0; T[0][2] = 0; T[0][3] = 0;
    T[1][0] = 0; T[1][1] = 1; T[1][2] = 0; T[1][3] = 0;
    T[2][0] = 0; T[2][1] = 0; T[2][2] = 1; T[2][3] = 0;
    T[3][0] = 0; T[3][1] = 0; T[3][2] = 0; T[3][3] = 1;
    return *this;
}

Projection3 &
Projection3::zero() {
    T[0][0] = 0; T[0][1] = 0; T[0][2] = 0; T[0][3] = 0;
    T[1][0] = 0; T[1][1] = 0; T[1][2] = 0; T[1][3] = 0;
    T[2][0] = 0; T[2][1] = 0; T[2][2] = 0; T[2][3] = 0;
    T[3][0] = 0; T[3][1] = 0; T[3][2] = 0; T[3][3] = 0;
    return *this;
}

Projection3 &
Projection3::perspective(double fov, double aspect, 
			 double znear, double zfar) {
    return *this *= ::perspective(fov, aspect, znear, zfar);
}

Projection3 &
Projection3::frustum(double l, double r, double t,
		     double b, double n, double f) {
    return *this *= ::frustum(l, r, t, b, n, f);
}

Projection3 &
Projection3::orthox() {
    return *this *= ::orthox();
}

Projection3 &
Projection3::orthoy() {
    return *this *= ::orthoy();
}

Projection3 &
Projection3::orthoz() {
    return *this *= ::orthoz();
}

Projection3 &
Projection3::scale(double sx, double sy, double sz) {
    return *this *= ::scale(sx, sy, sz);
}

Projection3 &
Projection3::translate(double tx, double ty, double tz) {
    return *this *= ::translate(tx, ty, tz);
}

Projection3 &
Projection3::rotatex(double theta) {
    return *this *= ::rotatex(theta);
}

Projection3 &
Projection3::rotatey(double theta) {
    return *this *= ::rotatey(theta);
}

Projection3 &
Projection3::rotatez(double theta) {
    return *this *= ::rotatez(theta);
}

Projection3 &
Projection3::rotate(double theta, const Vector3 &axis) {
    return *this *= ::rotate(theta, axis);
}

Projection3 &
Projection3::transpose() {
    swap(T[0][1], T[1][0]);
    swap(T[0][2], T[2][0]);
    swap(T[0][3], T[3][0]);
    swap(T[1][2], T[2][1]);
    swap(T[1][3], T[3][1]);
    swap(T[2][3], T[3][2]);
    return *this;
}

Projection3 &
Projection3::invert() {
    double d = determinant();
    assert(d != 0);
    adjoint();
    for (int i = 0; i < 4; ++i)
	for (int j = 0; j < 4; ++j)
	    T[i][j] /= d;
    return *this;
}

double
Projection3::determinant() const {
    double ans;
    double a1, a2, a3, a4, b1, b2, b3, b4, c1, c2, c3, c4, d1, d2, d3, d4;

    /* assign to individual variable names to aid selecting */
    /*  correct elements */

    a1 = T[0][0]; b1 = T[0][1]; 
    c1 = T[0][2]; d1 = T[0][3];

    a2 = T[1][0]; b2 = T[1][1]; 
    c2 = T[1][2]; d2 = T[1][3];

    a3 = T[2][0]; b3 = T[2][1]; 
    c3 = T[2][2]; d3 = T[2][3];

    a4 = T[3][0]; b4 = T[3][1]; 
    c4 = T[3][2]; d4 = T[3][3];

    ans = a1 * det3x3(b2, b3, b4, c2, c3, c4, d2, d3, d4)
        - b1 * det3x3(a2, a3, a4, c2, c3, c4, d2, d3, d4)
        + c1 * det3x3(a2, a3, a4, b2, b3, b4, d2, d3, d4)
        - d1 * det3x3(a2, a3, a4, b2, b3, b4, c2, c3, c4);

    return ans;
}

double
Projection3::det3x3(double a1, double a2, double a3, 
		    double b1, double b2, double b3,
		    double c1, double c2, double c3) {
    return (a1 * (b2 * c3 - b3 * c2)
	    - b1 * (a2 * c3 - a3 * c2)
	    + c1 * (a2 * b3 - a3 * b2));
}

Projection3 &
Projection3::adjoint() {
    double a1, a2, a3, a4, b1, b2, b3, b4;
    double c1, c2, c3, c4, d1, d2, d3, d4;

    /* assign to individual variable names to aid  */
    /* selecting correct values  */

    a1 = T[0][0]; b1 = T[0][1]; 
    c1 = T[0][2]; d1 = T[0][3];

    a2 = T[1][0]; b2 = T[1][1]; 
    c2 = T[1][2]; d2 = T[1][3];

    a3 = T[2][0]; b3 = T[2][1];
    c3 = T[2][2]; d3 = T[2][3];

    a4 = T[3][0]; b4 = T[3][1]; 
    c4 = T[3][2]; d4 = T[3][3];

    /* row column labeling reversed since we transpose rows & columns */

    T[0][0]  =   det3x3(b2, b3, b4, c2, c3, c4, d2, d3, d4);
    T[1][0]  = - det3x3(a2, a3, a4, c2, c3, c4, d2, d3, d4);
    T[2][0]  =   det3x3(a2, a3, a4, b2, b3, b4, d2, d3, d4);
    T[3][0]  = - det3x3(a2, a3, a4, b2, b3, b4, c2, c3, c4);
        
    T[0][1]  = - det3x3(b1, b3, b4, c1, c3, c4, d1, d3, d4);
    T[1][1]  =   det3x3(a1, a3, a4, c1, c3, c4, d1, d3, d4);
    T[2][1]  = - det3x3(a1, a3, a4, b1, b3, b4, d1, d3, d4);
    T[3][1]  =   det3x3(a1, a3, a4, b1, b3, b4, c1, c3, c4);
        
    T[0][2]  =   det3x3(b1, b2, b4, c1, c2, c4, d1, d2, d4);
    T[1][2]  = - det3x3(a1, a2, a4, c1, c2, c4, d1, d2, d4);
    T[2][2]  =   det3x3(a1, a2, a4, b1, b2, b4, d1, d2, d4);
    T[3][2]  = - det3x3(a1, a2, a4, b1, b2, b4, c1, c2, c4);
        
    T[0][3]  = - det3x3(b1, b2, b3, c1, c2, c3, d1, d2, d3);
    T[1][3]  =   det3x3(a1, a2, a3, c1, c2, c3, d1, d2, d3);
    T[2][3]  = - det3x3(a1, a2, a3, b1, b2, b3, d1, d2, d3);
    T[3][3]  =   det3x3(a1, a2, a3, b1, b2, b3, c1, c2, c3);

    return *this;
}

Point3 
Projection3::project(const Point3 &p) const {
    double w = T[3][0] * p.x + T[3][1] * p.y + T[3][2] * p.z + T[3][3];
    assert(w != 0);
    Point3 ret(T[0][0] * p.x + T[0][1] * p.y + T[0][2] * p.z + T[0][3],
	       T[1][0] * p.x + T[1][1] * p.y + T[1][2] * p.z + T[1][3],
	       T[2][0] * p.x + T[2][1] * p.y + T[2][2] * p.z + T[2][3]);
    ret.x /= w; ret.y /= w; ret.z /= w;
    return ret;
}

Vector3 
Projection3::project(const Vector3 &v) const {
#if 0
    double w = T[3][0] * v.x + T[3][1] * v.y + T[3][2] * v.z + T[3][3];
    assert(w != 0);
    Vector3 ret(T[0][0] * v.x + T[0][1] * v.y + T[0][2] * v.z + T[0][3],
		T[1][0] * v.x + T[1][1] * v.y + T[1][2] * v.z + T[1][3],
		T[2][0] * v.x + T[2][1] * v.y + T[2][2] * v.z + T[2][3]);
    return ret /= w;
#endif
    return Vector3(T[0][0] * v.x + T[0][1] * v.y + T[0][2] * v.z,
		   T[1][0] * v.x + T[1][1] * v.y + T[1][2] * v.z,
		   T[2][0] * v.x + T[2][1] * v.y + T[2][2] * v.z);
}

Normal3
Projection3::project(const Normal3 &n, const Projection3 *inv) const {
#if 0
    if (inv) {
	// skip transpose and use rows instead of columns
	Normal3 ret(inv->T[0][0] * n.x + inv->T[1][0] * n.y + inv->T[2][0] * n.z +
		    inv->T[3][0],
		    inv->T[0][1] * n.x + inv->T[1][1] * n.y + inv->T[2][1] * n.z +
		    inv->T[3][1],
		    inv->T[0][2] * n.x + inv->T[1][2] * n.y + inv->T[2][2] * n.z +
		    inv->T[3][2]);
	double w = inv->T[0][3] * n.x + inv->T[1][3] * n.y + inv->T[2][3] * n.z + 
	    inv->T[3][3];
	assert(w != 0);
	return ret /= w;
    }
    else {
	Projection3 inv1 = inverse();
	Normal3 ret(inv1.T[0][0] * n.x + inv1.T[1][0] * n.y + inv1.T[2][0] * n.z +
		    inv1.T[3][0],
		    inv1.T[0][1] * n.x + inv1.T[1][1] * n.y + inv1.T[2][1] * n.z +
		    inv1.T[3][1],
		    inv1.T[0][2] * n.x + inv1.T[1][2] * n.y + inv1.T[2][2] * n.z +
		    inv1.T[3][2]);
	double w = inv1.T[0][3] * n.x + inv1.T[1][3] * n.y + inv1.T[2][3] * n.z +
	    inv1.T[3][3];
	assert(w != 0);
	return ret /= w;
    }
#endif
    if (inv) {
	// skip transpose and use rows instead of columns
	return Normal3(inv->T[0][0] * n.x + inv->T[1][0] * n.y + inv->T[2][0] * n.z, 
		       inv->T[0][1] * n.x + inv->T[1][1] * n.y + inv->T[2][1] * n.z,
		       inv->T[0][2] * n.x + inv->T[1][2] * n.y + inv->T[2][2] * n.z);
    }
    else {
	Projection3 inv1 = inverse();
	return Normal3(inv1.T[0][0] * n.x + inv1.T[1][0] * n.y + inv1.T[2][0] * n.z,
		       inv1.T[0][1] * n.x + inv1.T[1][1] * n.y + inv1.T[2][1] * n.z,
		       inv1.T[0][2] * n.x + inv1.T[1][2] * n.y + inv1.T[2][2] * n.z);
    }
}

Homogeneous3
Projection3::project(const Homogeneous3 &h) const {
    return Homogeneous3(h.x * T[0][0] + h.y * T[0][1] + 
			    h.z * T[0][2] + h.d * T[0][3],
			h.x * T[1][0] + h.y * T[1][1] + 
			    h.z * T[1][2] + h.d * T[1][3],
			h.x * T[2][0] + h.y * T[2][1] + 
			    h.z * T[2][2] + h.d * T[2][3],
			h.x * T[3][0] + h.y * T[3][1] + 
			    h.z * T[3][2] + h.d * T[3][3]);
}

bool
Projection3::operator==(const Projection3 &p) const {
    for (int i = 0; i < 4; ++i)
	for (int j = 0; j < 4; ++j)
	    if (T[i][j] != p.T[i][j])
		return false;
    return true;
}

bool
Projection3::operator!=(const Projection3 &p) const {
    for (int i = 0; i < 4; ++i)
	for (int j = 0; j < 4; ++j)
	    if (T[i][j] != p.T[i][j])
		return true;
    return false;
}

std::ostream &
operator<<(std::ostream &os, const Projection3 &p) {
    for (int i = 0; i < 4; ++i) {
	os << '[';
	for (int j = 0; j < 4; ++j) {
	    os << p.T[i][j];
	    if (j != 3)
		os << ' ';
	    else
		os << ']' << std::endl;
	}
    }
    return os;
}

Projection3 identityP3() {
    return Projection3(1, 0, 0, 0,
		       0, 1, 0, 0,
		       0, 0, 1, 0,
		       0, 0, 0, 1);
}

Projection3
perspective(double fov, double aspect, double znear, double zfar) {
    double cotfov = tan(fov / 2.0);
    if(cotfov != 0.0) {
	cotfov = 1.0 / cotfov;
    }
    return Projection3(
       cotfov / aspect, 0,       0,                      0,
       0,               cotfov,  0,                      0,
       0,               0,      -(zfar+znear)/(zfar-znear), -2*zfar*znear/(zfar-znear),
       0,               0,      -1,                      0);
}

Projection3
frustum(double l, double r, double t,
	double b, double n, double f) {
    double dx = r - l;
    double dy = t - b;
    double dz = f - n;

    return Projection3(2*n / dx,  0,         (r+l)/dx,  0,
		       0,         2*n / dy,  (t+b)/dy,  0,
		       0,         0,        -(f+n)/dz,  -(2*f*n)/dz,
		       0,         0,        -1,         0);
}

Projection3
orthox() {
    return Projection3(0, 0, 0, 0,
		       0, 1, 0, 0,
		       0, 0, 1, 0,
		       0, 0, 0, 1);
}
		       
Projection3
orthoy() {
    return Projection3(1, 0, 0, 0,
		       0, 0, 0, 0,
		       0, 0, 1, 0, 
		       0, 0, 0, 1);
}

Projection3
orthoz() {
    return Projection3(1, 0, 0, 0,
		       0, 1, 0, 0,
		       0, 0, 0, 0,
		       0, 0, 0, 1);
}
