#include <math.h>
#include "Matrix.h"
#include "Random.h"

Matrix &
Matrix::identity(void)
{
  assert(r==c);
  for (int i=0; i<r; i++) 
    for (int j=0; j<c; j++) 
      data[i][j] = (i==j ? 1.0 : 0.0);
  return *this;
}

Matrix &
Matrix::ones(void)
{
  for (int i=0; i<r; i++) 
    for (int j=0; j<c; j++) 
      data[i][j] = 1.0;
  return *this;
}

Matrix &
Matrix::zero(void)
{
  for (int i=0; i<r; i++) 
    for (int j=0; j<c; j++) 
      data[i][j] = 0.0;
  return *this;
}

#define SWAP(a,b) { temp=(a); (a)=(b); (b)=temp; }

Matrix &
Matrix::invert(void)
{
  assert(r==c);
  // Gauss Jordan elimination, adapted from Num.Recipes
  int *indxc,*indxr,*ipiv;
  int i,icol,irow,j,k,l,ll;
  double big,dum,pivinv,temp;

  indxc = new int[r];
  indxr = new int[r];
  ipiv  = new int[r];
  for (j=0;j<r;j++) ipiv[j]=0;
  for (i=0;i<r;i++) {
    big=0.0;
    for (j=0;j<r;j++)
      if (ipiv[j] != 1)
	for (k=0;k<r;k++) {
	  if (ipiv[k] == 0) {
	    if (fabs(data[j][k]) >= big) {
	      big=fabs(data[j][k]);
	      irow=j;
	      icol=k;
	    }
	  } else if (ipiv[k] > 1) {
	    cerr << "gaussj: Singular Matrix-1" << endl;
	    return *this;
	  }
	}
    ++(ipiv[icol]);
    if (irow != icol) {
      for (l=0;l<r;l++) SWAP(data[irow][l],data[icol][l]);
//      for (l=0;l<m;l++) SWAP(b[irow][l],b[icol][l]);
    }
    indxr[i]=irow;
    indxc[i]=icol;
    if (data[icol][icol] == 0.0) {
      cerr << "gaussj: Singular Matrix-2" << endl;
      return *this;
    }
    pivinv=1.0/data[icol][icol];
    data[icol][icol]=1.0;
    for (l=0;l<r;l++) data[icol][l] *= pivinv;
//    for (l=0;l<m;l++) b[icol][l] *= pivinv;
    for (ll=0;ll<r;ll++)
      if (ll != icol) {
	dum=data[ll][icol];
	data[ll][icol]=0.0;
	for (l=0;l<r;l++) data[ll][l] -= data[icol][l]*dum;
//	for (l=0;l<m;l++) b[ll][l] -= b[icol][l]*dum;
      }
  }
  for (l=r-1;l>=0;l--) {
    if (indxr[l] != indxc[l])
      for (k=0;k<r;k++)
	SWAP(data[k][indxr[l]],data[k][indxc[l]]);
  }

  delete[] indxc;
  delete[] indxr;
  delete[] ipiv;
  return *this;
}

double &
Matrix::operator()(int i, int j)
{
  return data[i][j];
}

Matrix
Matrix::transpose(void)
{
  Matrix res(c,r);
  for (int i=0; i<r; i++)
    for (int j=0; j<c; j++)
      res.data[j][i] = data[i][j];
  return res;
}

Matrix &
Matrix::operator +=(const Matrix &a)
{
  assert(a.c == c && a.r == r);
  for (int i=0; i<r; i++)
    for (int j=0; j<c; j++) {
      data[i][j] += a.data[i][j];
    }
  return *this;
}

Matrix
operator *(const Matrix &a, const Matrix &b)
{
  assert(a.c == b.r);
  Matrix res(a.r, b.c);
  for (int i=0; i<res.r; i++)
    for (int j=0; j<res.c; j++) {
      res.data[i][j] = 0.0;
      for (int k=0; k<a.c; k++)
	res.data[i][j] += a.data[i][k] * b.data[k][j];
    }
  return res;
}

Matrix
Tmult(const Matrix &a, const Matrix &b)
{
  assert(a.r == b.r);
  Matrix res(a.c, b.c);
  for (int i=0; i<res.r; i++)
    for (int j=0; j<res.c; j++) {
      res.data[i][j] = 0.0;
      for (int k=0; k<a.r; k++)
	res.data[i][j] += a.data[k][i] * b.data[k][j];
    }
  return res;
}

void 
mult(const Matrix &a, double *b, double *c)
{
  // c = A * b
  for (int i=0; i<a.r; i++) {
    c[i] = 0.0;
    for (int j=0; j<a.c; j++) {
      c[i] += a.data[i][j] * b[j];
    }
  }
}

void 
Tmult(const Matrix &a, double *b, double *c)
{
  // c = At * b
  for (int i=0; i<a.c; i++) {
    c[i] = 0.0;
    for (int j=0; j<a.r; j++) {
      c[i] += a.data[j][i] * b[j];
    }
  }
}

void 
square(const Matrix &a, Matrix &ata)
{
  // ata = At * A
  int i,j;
  // calculate upper triangle
  for (i=0; i<a.c; i++)
    for (j=i; j<a.c; j++) {
      ata.data[i][j] = 0.0;
      for (int k=0; k<a.r; k++)
	ata.data[i][j] += a.data[k][i] * a.data[k][j];
      // copy to lower triangle
      ata.data[j][i] = ata.data[i][j];
    }
}

ostream &
operator <<(ostream &os, const Matrix &m)
{
  for (int i=0; i<m.r; i++) {
    for (int j=0; j<m.c; j++) {
      cout << m.data[i][j] << " ";
    }
    os << endl;
  }
  return os;
}

int
gauss_jordan(Matrix &a, Matrix &b)
{
  int r = a.r;
  int m = b.c;
  assert(a.c==a.r && a.c==b.r);
  // Gauss Jordan elimination, adapted from Num.Recipes
  int *indxc,*indxr,*ipiv;
  int i,icol,irow,j,k,l,ll;
  double big,dum,pivinv,temp;

  indxc = new int[r];
  indxr = new int[r];
  ipiv  = new int[r];
  for (j=0;j<r;j++) ipiv[j]=0;
  for (i=0;i<r;i++) {
    big=0.0;
    for (j=0;j<r;j++)
      if (ipiv[j] != 1)
	for (k=0;k<r;k++) {
	  if (ipiv[k] == 0) {
	    if (fabs(a.data[j][k]) >= big) {
	      big=fabs(a.data[j][k]);
	      irow=j;
	      icol=k;
	    }
	  } else if (ipiv[k] > 1) {
	    // cerr << "gaussj: Singular Matrix-1" << endl;
	    return 0;
	  }
	}
    ++(ipiv[icol]);
    if (irow != icol) {
      for (l=0;l<r;l++) SWAP(a.data[irow][l],a.data[icol][l]);
      for (l=0;l<m;l++) SWAP(b.data[irow][l],b.data[icol][l]);
    }
    indxr[i]=irow;
    indxc[i]=icol;
    if (a.data[icol][icol] == 0.0) {
      // cerr << "gaussj: Singular Matrix-2" << endl;
      return 0;
    }
    pivinv=1.0/a.data[icol][icol];
    a.data[icol][icol]=1.0;
    for (l=0;l<r;l++) a.data[icol][l] *= pivinv;
    for (l=0;l<m;l++) b.data[icol][l] *= pivinv;
    for (ll=0;ll<r;ll++)
      if (ll != icol) {
	dum=a.data[ll][icol];
	a.data[ll][icol]=0.0;
	for (l=0;l<r;l++) a.data[ll][l] -= a.data[icol][l]*dum;
	for (l=0;l<m;l++) b.data[ll][l] -= b.data[icol][l]*dum;
      }
  }
  for (l=r-1;l>=0;l--) {
    if (indxr[l] != indxc[l])
      for (k=0;k<r;k++)
	SWAP(a.data[k][indxr[l]],a.data[k][indxc[l]]);
  }

  delete[] indxc;
  delete[] indxr;
  delete[] ipiv;
  return 1;
}


int
gauss_jordan_fast(Matrix &a, double *b)
{
  int r = a.r;
  // Gauss Jordan elimination, adapted from Num.Recipes
  int *ipiv;
  int i,icol,irow,j,k,l,ll;
  double big,dum,pivinv,temp;

  ipiv  = new int[r];
  for (j=0;j<r;j++) ipiv[j]=0;
  for (i=0;i<r;i++) {
    big=0.0;
    for (j=0;j<r;j++)
      if (ipiv[j] != 1)
	for (k=0;k<r;k++) {
	  if (ipiv[k] == 0) {
	    if (fabs(a.data[j][k]) >= big) {
	      big=fabs(a.data[j][k]);
	      irow=j;
	      icol=k;
	    }
	  } else if (ipiv[k] > 1) {
	    // cerr << "gaussj: Singular Matrix-1" << endl;
	    return 0;
	  }
	}
    ++(ipiv[icol]);
    if (irow != icol) {
      for (l=0;l<r;l++) SWAP(a.data[irow][l],a.data[icol][l]);
      SWAP(b[irow],b[icol]);
    }
    if (a.data[icol][icol] == 0.0) {
      // cerr << "gaussj: Singular Matrix-2" << endl;
      return 0;
    }
    pivinv=1.0/a.data[icol][icol];
    a.data[icol][icol]=1.0;
    for (l=0;l<r;l++) a.data[icol][l] *= pivinv;
    b[icol] *= pivinv;
    for (ll=0;ll<r;ll++)
      if (ll != icol) {
	dum=a.data[ll][icol];
	a.data[ll][icol]=0.0;
	for (l=0;l<r;l++) a.data[ll][l] -= a.data[icol][l]*dum;
	b[ll] -= b[icol]*dum;
      }
  }
  delete[] ipiv;
  return 1;
}

void
Matrix::HouseholderQR(void)
{
  int i,j,k;
  double *v = new double[r];
  double *w = new double[c];
  double mu;
  double beta;
  for (j=0; j<c; j++) {
    // v(j:m) = house(A(j:m,j))
    {
      // mu = ||x||_2
      mu = 0.0;
      for (i=j; i<r; i++) mu += data[i][j]*data[i][j];
      mu = sqrt(mu);
      if (mu != 0.0) {
	// beta = x(1) + sign(x(1)) mu
	beta = data[j][j];
	if (beta>0.0) beta += mu;
	else          beta -= mu;
	for (i=j+1; i<r; i++) v[i] = data[i][j]/beta;
      } else {
	for (i=j+1; i<r; i++) v[i] = data[i][j];
      }
      v[j] = 1.0;
    }
    // A(j:m,j:n) = row.house(A(j:m,j:n),v(j:m))
    {
      // beta = -2/(vtv)
      mu = 0.0;
      for (i=j; i<r; i++) mu += v[i]*v[i];
      beta = -2.0/mu;
      // w = beta At v
      for (i=j; i<c; i++) {
	w[i] = 0.0;
	for (k=j; k<r; k++) w[i] += data[k][i] * v[k];
	w[i] *= beta;
      }
      // A = A + v.wt
      for (i=j; i<r; i++)
	for (k=j; k<c; k++)
	  data[i][k] += v[i] * w[k];
    }
    for (i=j+1; i<r; i++)
      data[i][j] = v[i];
  }
  delete[] v;
  delete[] w;
}

void
Matrix::HouseholderLSQ(double *b)
{
  int i,j;
  // overwrite A with its QR factorization
  HouseholderQR();
  for (j=0; j<c; j++) {
    // apply the j-th Householder transformation
    double vtv = 1.0;
    for (i=j+1; i<r; i++) vtv += data[i][j] * data[i][j];
    double beta = -2.0 / vtv;
    double w = b[j];
    for (i=j+1; i<r; i++) w += b[i] * data[i][j];
    w *= beta;
    b[j] += w;
    for (i=j+1; i<r; i++) b[i] += data[i][j] * w;
  }
  // solve R(1:n,1:n) xLS = b(1:n) using back substitution
  for (j=c-1; j>=0; j--) {
    b[j] /= data[j][j];
    for (i=0; i<j; i++) {
      b[i] -= data[i][j] * b[j];
    }
  }
}


double
Matrix::maxEig(void)
{
  double *a = new double[r];
  double *b = new double[r];
  for (int i=0; i<r; i++) a[i] = 1.0;

  double max;
  for (int j=0; j<5; j++) {
    mult(*this, a, b);
    mult(*this, b, a);
    // normalize (Linf norm)
    max = a[0];
	int i;
    for ( i=1; i<r; i++) if (max < a[i]) max = a[i];
    for ( i=0; i<r; i++) a[i] /= max;
  }
  delete[] a;
  delete[] b;

  return sqrt(max);
}


double
Matrix::minEig(void)
{
  Matrix inv = *this;
  inv.invert();
  return 1.0/inv.maxEig();
}


LSQ::LSQ(int i, int j)
{
  AtA = new Matrix(i,i);
  AtB = new Matrix(i,j);
  ac = i;
  bc = j;
  clear();
}

LSQ::~LSQ(void)
{
  delete AtA; delete AtB;
}

void
LSQ::clear(void)
{
  AtA->zero();
  AtB->zero();
}

void
LSQ::add(const double *A, const double *B)
{
  // fill only the upper triangle of AtA
  for (int i=0; i<ac; i++) {
    (*AtA)(i,i) += A[i]*A[i];
    for (int j=i+1; j<ac; j++) {
      (*AtA)(i,j) += A[i]*A[j];
    }
    {for (int j=0; j<bc; j++) {
      (*AtB)(i,j) += A[i]*B[j];
    }}
  }
}

void
LSQ::add(const float *A, const float *B)
{
  // fill only the upper triangle of AtA
  for (int i=0; i<ac; i++) {
    (*AtA)(i,i) += A[i]*A[i];
    for (int j=i+1; j<ac; j++) {
      (*AtA)(i,j) += A[i]*A[j];
    }
	{
    for (int j=0; j<bc; j++) {
      (*AtB)(i,j) += A[i]*B[j];
    }
	}
  }
}

Matrix &
LSQ::solve(void)
{
  // copy the upper triangle of AtA to lower
  for (int i=1; i<ac; i++)
    for (int j=0; j<i; j++)
      (*AtA)(i,j) = (*AtA)(j,i);
  gauss_jordan(*AtA, *AtB);
  return *AtB;
}


//i don't need or want main here...so it is gone
//-alex
#if 0

#define SHOW(x) cout << #x " = " << x << endl

int
main()
{
  Matrix m(3,3);
  m.identity();
  cout << m;
  m(0,2) = 6.0;
  cout << m;
  cout << m.transpose();

  Matrix n(3,3); n=m; cout << n; 
  Matrix a(3,2);
  a(0,0) = 1.0; a(0,1) = 1.0;
  a(1,0) = 2.0; a(1,1) = 1.0;
  a(2,0) = 4.0; a(2,1) = 1.0;
  cout << a;
  cout << m*a;

  Matrix b = a.transpose() * a;
  cout << b;
  Matrix bi = b;
  bi.invert();
  cout << bi;
  cout << b*bi;

  LSQ lsq(3,2);
  double A[3], Arot[3];
  Matrix R(3,3);
  R(0,0) = 0.988538;    R(0,1) = -0.103803; R(0,2) = 0.109629;
  R(1,0) = 0.150823;    R(1,1) = 0.711603;  R(1,2) = -0.686203;
  R(2,0) = -0.00678231; R(2,1) = 0.694872;  R(2,2) = 0.719102;

  Random rnd(0);

  cout << R;
  for (int i=0; i<40; i++) {
    A[0] = rnd();
    A[1] = rnd();
    A[2] = rnd();
    Tmult(R,A,Arot);
    lsq.add(A,Arot);
  }
  cout << lsq.solve();

  cout << "testing QR" << endl;
  
  Matrix qA(10,3),qa(10,3);
  double qb[10],qc[10];
  for (i=0; i<10; i++) {
    qA(i,0) = rnd();
    qA(i,1) = rnd();
    qA(i,2) = rnd();
    qb[i]   = 1*qA(i,0) + 2*qA(i,1) + 3*qA(i,2);
  }
  qa = qA;
  cout << qa;
  for (i=0; i<10; i++) cout << qb[i] << " "; cout << endl;
  qa.HouseholderLSQ(qb);
  for (i=0; i<10; i++) cout << qb[i] << " "; cout << endl;
  cout << qA;
  mult(qA,qb,qc);
  for (i=0; i<10; i++) cout << qc[i] << " "; cout << endl;

  cout << "Testing eigenvalues" << endl;
  double me[9] = { 1,2,3,2,3,4,3,4,5 };
  R.set(me);
  SHOW(R);
  cout << "Matlab gives: 0.0000 -0.6235 9.6235" << endl;
  SHOW(R.maxEig());
  SHOW(R.minEig());
}
#endif
